import datetime
import typing

import tzlocal
from pytz import utc

import HABApp
from .base_item_times import BaseWatch, ChangedTime, UpdatedTime
from .tmp_data import add_tmp_data as _add_tmp_data
from .tmp_data import restore_tmp_data as _restore_tmp_data


class BaseItem:
    """BaseItem, all items must inherit from this class
    """

    @classmethod
    def get_item(cls, name: str):
        """Returns an already existing item. If it does not exist or has a different item type an exception will occur.

        :param name: Name of the item
        """
        assert isinstance(name, str), type(name)
        item = HABApp.core.Items.get_item(name)
        assert isinstance(item, cls), f'{cls} != {type(item)}'
        return item

    def __init__(self, name: str):
        super().__init__()
        assert isinstance(name, str), type(name)

        self._name: str = name

        _now = datetime.datetime.now(tz=utc)
        self._last_change: ChangedTime = ChangedTime(self._name, _now)
        self._last_update: UpdatedTime = UpdatedTime(self._name, _now)

    @property
    def name(self) -> str:
        """
        :return: Name of the item (read only)
        """
        return self._name

    @property
    def last_change(self) -> datetime.datetime:
        """
        :return: Timestamp of the last time when the item has been changed (read only)
        """
        return self._last_change.dt.astimezone(tzlocal.get_localzone()).replace(tzinfo=None)

    @property
    def last_update(self) -> datetime.datetime:
        """
        :return: Timestamp of the last time when the item has been updated (read only)
        """
        return self._last_update.dt.astimezone(tzlocal.get_localzone()).replace(tzinfo=None)

    def __repr__(self):
        ret = ''
        for k in ['name', 'last_change', 'last_update']:
            ret += f'{", " if ret else ""}{k}: {getattr(self, k)}'
        return f'<{self.__class__.__name__} {ret:s}>'

    def watch_change(self, secs: typing.Union[int, float]) -> BaseWatch:
        """Generate an event if the item does not change for a certain period of time.
        Has to be called from inside a rule function.

        :param secs: secs after which the event will occur, max 1 decimal digit for floats
        :return: The watch obj which can be used to cancel the watch
        """
        if isinstance(secs, float):
            secs = round(secs, 1)
        else:
            assert isinstance(secs, int)
        assert secs > 0, secs
        w = self._last_change.add_watch(secs)
        HABApp.rule.get_parent_rule().register_cancel_obj(w)
        return w

    def watch_update(self, secs: typing.Union[int, float]) -> BaseWatch:
        """Generate an event if the item does not receive and update for a certain period of time.
        Has to be called from inside a rule function.

        :param secs: secs after which the event will occur, max 1 decimal digit for floats
        :return: The watch obj which can be used to cancel the watch
        """
        if isinstance(secs, float):
            secs = round(secs, 1)
        else:
            assert isinstance(secs, int)
        assert secs > 0, secs
        w = self._last_update.add_watch(secs)
        HABApp.rule.get_parent_rule().register_cancel_obj(w)
        return w

    def listen_event(self, callback: typing.Callable[[typing.Any], typing.Any],
                     event_type: typing.Union['HABApp.core.events.AllEvents', typing.Any]
                     ) -> 'HABApp.core.EventBusListener':
        """
        Register an event listener which listens to all event that the item receives

        :param callback: callback that accepts one parameter which will contain the event
        :param event_type: Event filter. This is typically :class:`~HABApp.core.ValueUpdateEvent` or
            :class:`~HABApp.core.ValueChangeEvent` which will also trigger on changes/update from openHAB
            or mqtt.
        """
        rule = HABApp.rule.get_parent_rule()
        return rule.listen_event(self._name, callback=callback, event_type=event_type)

    def _on_item_add(self):
        """This function gets automatically called when the item is added to the item registry
        """
        _restore_tmp_data(self)

    def _on_item_remove(self):
        """This function gets automatically called when the item is removed from the item registry
        """
        _add_tmp_data(self)



from typing import Set, Dict

import HABApp
from HABApp.openhab.connection_handler.func_async import async_set_habapp_metadata, async_create_item, \
    async_remove_item, async_create_channel_link, async_get_items, \
    async_remove_channel_link, async_remove_metadata, async_set_metadata, async_get_item_with_habapp_meta
from HABApp.openhab.definitions.rest.habapp_data import HABAppThingPluginData, load_habapp_meta
from ._log import log_item as log
from .cfg_validator import UserItem


def _filter_items(i):
    if not i.get('editable', False):
        return False

    if 'HABApp' not in i.setdefault('metadata', {}):
        return False

    load_habapp_meta(i)
    if not isinstance(i['metadata']['HABApp'], HABAppThingPluginData):
        return False
    return True


async def cleanup_items(keep_items: Set[str]):
    all_items = await async_get_items(include_habapp_meta=True)

    to_delete: Dict[str, HABAppThingPluginData] = {}
    for cfg in filter(_filter_items, all_items):
        name = cfg['name']
        if name not in keep_items:
            to_delete[name] = cfg['metadata']['HABApp']

    if not to_delete:
        return None

    for item, data in to_delete.items():
        assert isinstance(data, HABAppThingPluginData)
        await _remove_item(item, data)


async def _remove_item(item: str, data: HABAppThingPluginData):
    # remove created link
    if data.created_link is not None:
        log.debug(f'Removing link from {data.created_link} to {item}')
        await async_remove_channel_link(data.created_link, item)

    # remove created metadata
    for ns in data.created_ns:
        log.debug(f'Removing metadata {ns} from {item}')
        await async_remove_metadata(item, ns)

    # finally remove the item
    log.info(f'Removing item {item}')
    await async_remove_item(item)


async def create_item(item: UserItem, test: bool) -> bool:

    if test:
        _txt = str(item)
        if _txt.startswith('UserItem'):
            _txt = _txt[4:]
        log.info(f'Would create {_txt}')
        return False

    name = item.name

    try:
        existing_ok = True
        existing_item = await async_get_item_with_habapp_meta(name)
        habapp_data = existing_item['metadata']['HABApp']

        # we only modify items we created
        if not isinstance(habapp_data, HABAppThingPluginData):
            log.warning(f'Skipping item {name} because it does already exist and was not created by the plugin!')
            return False

        # check if the item properties are already correct
        for k, v in item.get_oh_cfg().items():
            if v != existing_item.get(k, ''):
                existing_ok = False

    except HABApp.openhab.exceptions.ItemNotFoundError:
        existing_ok = True
        existing_item = None
        habapp_data = HABAppThingPluginData()

    # Update/Create item definition
    if existing_item is None or not existing_ok:
        log.info(f'{"Creating" if existing_item is None else "Updating"} item: {item.type} {name} "{item.label}"')
        tmp = item.get_oh_cfg()
        tmp['item_type'] = tmp.pop('type')
        if not await async_create_item(**tmp):
            log.error(f'Item operation failed for {tmp}!')
            return False
        await async_set_habapp_metadata(name, habapp_data)
    else:
        log.debug(f'Item {name} is already correct!')

    # check create link
    if item.link != habapp_data.created_link:
        # remove existing
        if habapp_data.created_link:
            log.debug(f'Removing link from {habapp_data.created_link} to {name}')
            await async_remove_channel_link(habapp_data.created_link, name)

        # create new link
        log.debug(f'Creating link from {item.link} to {item.name}')
        if not await async_create_channel_link(item.link, name):
            log.error(f'Creating link from {item.link} to {name} failed!')
            await _remove_item(name, habapp_data)
            return False

        # save that we created a link
        habapp_data.created_link = item.link
        await async_set_habapp_metadata(name, habapp_data)
    else:
        log.debug(f'Link to {name} is already correct')

    # check create metadata
    if item.metadata or habapp_data.created_ns:
        # remove obsolete
        for ns in set(habapp_data.created_ns) - set(item.metadata.keys()):
            log.debug(f'Removing metadata {ns} from {name}')
            await async_remove_metadata(name, ns)

        # create new
        for ns, meta_cfg in item.metadata.items():
            m_val = meta_cfg['value']
            m_config = meta_cfg['config']
            log.debug(f'Adding metadata {ns} to {name}: {m_val} {m_config}')
            if await async_set_metadata(name, ns, m_val, m_config):
                habapp_data.created_ns.append(ns)
            else:
                log.error(f'Creating metadata for {name} failed!')
                await _remove_item(name, habapp_data)
                return False

        # save that we created metadata
        habapp_data.created_ns = list(item.metadata.keys())
        await async_set_habapp_metadata(name, habapp_data)

    return True



from HABApp.openhab.items.base_item import OpenhabItem
from HABApp.openhab.items.commands import OnOffCommand, PercentCommand
from ..definitions import OnOffValue, PercentValue


class DimmerItem(OpenhabItem, OnOffCommand, PercentCommand):

    def set_value(self, new_value) -> bool:

        if isinstance(new_value, OnOffValue):
            new_value = 100 if new_value.on else 0
        elif isinstance(new_value, PercentValue):
            new_value = new_value.value

        # Percent is 0 ... 100
        if isinstance(new_value, (int, float)):
            assert 0 <= new_value <= 100, new_value
        else:
            assert new_value is None, new_value

        return super().set_value(new_value)

    def __str__(self):
        return self.value

    def is_on(self) -> bool:
        """Test value against on-value"""
        return bool(self.value)

    def is_off(self) -> bool:
        """Test value against off-value"""
        return not bool(self.value)



import logging

import HABApp
from HABApp.core import Items
from HABApp.core.wrapper import ignore_exception
from HABApp.openhab.map_items import map_item
from ._plugin import OnConnectPlugin
from ..interface_async import async_get_items, async_get_things

log = logging.getLogger('HABApp.openhab.items')


class LoadAllOpenhabItems(OnConnectPlugin):

    @ignore_exception
    async def on_connect_function(self):
        data = await async_get_items()
        if data is None:
            return None

        found_items = len(data)
        for _dict in data:
            item_name = _dict['name']
            new_item = map_item(item_name, _dict['type'], _dict['state'])
            if new_item is None:
                continue

            try:
                # if the item already exists and it has the correct type just update its state
                # Since we load the items before we load the rules this should actually never happen
                existing_item = Items.get_item(item_name)   # type: HABApp.core.items.BaseValueItem
                if isinstance(existing_item, new_item.__class__):
                    existing_item.set_value(new_item.value)  # use the converted state from the new item here
                    new_item = existing_item
            except Items.ItemNotFoundException:
                pass

            # create new item or change item type
            Items.add_item(new_item)

        # remove items which are no longer available
        ist = set(Items.get_all_item_names())
        soll = {k['name'] for k in data}
        for k in ist - soll:
            if isinstance(Items.get_item(k), HABApp.openhab.items.OpenhabItem):
                Items.pop_item(k)

        log.info(f'Updated {found_items:d} Items')


        # try to update things, too
        data = await async_get_things()
        if data is None:
            return None

        Thing = HABApp.openhab.items.Thing
        for t_dict in data:
            name = t_dict['UID']
            try:
                thing = HABApp.core.Items.get_item(name)
                if not isinstance(thing, Thing):
                    log.warning(f'Item {name} has the wrong type ({type(thing)}), expected Thing')
                    thing = Thing(name)
            except HABApp.core.Items.ItemNotFoundException:
                thing = Thing(name)

            thing.status = t_dict['statusInfo']['status']
            HABApp.core.Items.add_item(thing)

        # remove things which were deleted
        ist = set(HABApp.core.Items.get_all_item_names())
        soll = {k['UID'] for k in data}
        for k in ist - soll:
            if isinstance(HABApp.core.Items.get_item(k), Thing):
                HABApp.core.Items.pop_item(k)

        log.info(f'Updated {len(data):d} Things')
        return None


PLUGIN_LOAD_ITEMS = LoadAllOpenhabItems.create_plugin()



import asyncio
import functools
import logging
import re
import sys
import typing
from logging import Logger
from pathlib import Path

import stackprinter

import HABApp

log = logging.getLogger('HABApp')


SUPPRESSED_PATHS = (
    re.compile(f'[/\\\\]{Path(__file__).name}$'),   # this file

    # rule file loader
    re.compile(r'[/\\]rule_file.py$'),
    re.compile(r'[/\\]runpy.py$'),

    # Worker functions
    re.compile(r'[/\\]wrappedfunction.py$'),

    # Don't print stack for used libraries
    re.compile(r'[/\\](site-packages|lib|python\d\.\d)[/\\]asyncio[/\\]'),
    re.compile(r'[/\\]site-packages[/\\]aiohttp[/\\]'),
    re.compile(r'[/\\]site-packages[/\\]voluptuous[/\\]'),
    re.compile(r'[/\\]site-packages[/\\]pydantic[/\\]'),
)

SKIP_TB = tuple(re.compile(k.pattern.replace('$', ', ')) for k in SUPPRESSED_PATHS)


def format_exception(e: typing.Union[Exception, typing.Tuple[typing.Any, typing.Any, typing.Any]]) -> typing.List[str]:
    tb = []
    skip = 0

    lines = stackprinter.format(e, line_wrap=0, truncate_vals=2000, suppressed_paths=SUPPRESSED_PATHS).splitlines()
    for i, line in enumerate(lines):
        if not skip:
            for s in SKIP_TB:
                if s.search(line):
                    # if it's just a two line traceback we skip it
                    if lines[i + 1].startswith('    ') and lines[i + 2].startswith('File'):
                        skip = 2
                        continue
        if skip:
            skip -= 1
            continue

        tb.append(line)

    return tb


def process_exception(func: typing.Union[typing.Callable, str], e: Exception,
                      do_print=False, logger: logging.Logger = log):
    # lines = traceback.format_exc().splitlines()
    # del lines[0:3]
    lines = format_exception(e)

    func_name = func if isinstance(func, str) else func.__name__

    # log exception, since it is unexpected we push it to stdout, too
    if do_print:
        print(f'Error {e} in {func_name}:')
    logger.error(f'Error {e} in {func_name}:')
    for line in lines:
        if do_print:
            print(line)
        logger.error(line)

    # send Error to internal event bus so we can reprocess it and notify the user
    HABApp.core.EventBus.post_event(
        HABApp.core.const.topics.ERRORS, HABApp.core.events.habapp_events.HABAppError(
            func_name=func_name, exception=e, traceback='\n'.join(lines)
        )
    )


def log_exception(func):
    # return async wrapper
    if asyncio.iscoroutinefunction(func) or asyncio.iscoroutine(func):
        @functools.wraps(func)
        async def a(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except asyncio.CancelledError:
                pass
            except Exception as e:
                process_exception(func, e, do_print=True)
                # re raise exception, since this is something we didn't anticipate
                raise

        return a

    @functools.wraps(func)
    def f(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            process_exception(func, e, do_print=True)
            # re raise exception, since this is something we didn't anticipate
            raise

    return f


def ignore_exception(func):
    # return async wrapper
    if asyncio.iscoroutinefunction(func) or asyncio.iscoroutine(func):
        @functools.wraps(func)
        async def a(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except asyncio.CancelledError:
                pass
            except Exception as e:
                process_exception(func, e)
                return None

        return a

    @functools.wraps(func)
    def f(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            process_exception(func, e)
            return None
    return f


class ExceptionToHABApp:
    def __init__(self, logger: typing.Optional[Logger] = None, log_level: int = logging.ERROR,
                 ignore_exception: bool = True):
        self.log: typing.Optional[Logger] = logger
        self.log_level = log_level
        self.ignore_exception: bool = ignore_exception

        self.raised_exception = False

        self.proc_tb: typing.Optional[typing.Callable[[list], list]] = None

    def __enter__(self):
        self.raised_exception = False

    def __exit__(self, exc_type, exc_val, exc_tb):
        # no exception -> we exit gracefully
        if exc_type is None and exc_val is None:
            return True

        self.raised_exception = True

        # tb = traceback.format_exception(exc_type, exc_val, exc_tb)
        # # there is an inconsistent use of newlines and array entries so we normalize it
        # tb = '\n'.join(map(lambda x: x.strip(' \n'), tb))
        # tb = tb.splitlines()

        tb = format_exception((exc_type, exc_val, exc_tb))

        # possibility to reprocess tb
        if self.proc_tb is not None:
            tb = self.proc_tb(tb)

        # try to get the parent function name
        try:
            f_name = sys._getframe().f_back.f_code.co_name
        except Exception:
            f_name = 'Exception while getting the function name!'

        # log error
        if self.log is not None:
            self.log.log(self.log_level, f'Error "{exc_val}" in {f_name}:')
            for line in tb:
                self.log.log(self.log_level, line)

        # send Error to internal event bus so we can reprocess it and notify the user
        HABApp.core.EventBus.post_event(
            HABApp.core.const.topics.WARNINGS if self.log_level == logging.WARNING else HABApp.core.const.topics.ERRORS,
            HABApp.core.events.habapp_events.HABAppError(
                func_name=f_name, exception=exc_val, traceback='\n'.join(tb)
            )
        )
        return self.ignore_exception



import typing
import HABApp.core

from ..map_values import map_openhab_values
from .base_event import OpenhabEvent

# smarthome/items/NAME/state -> 16
# openhab/items/NAME/state   -> 14
# todo: revert this once we go OH3 only
NAME_START: int = 14


class ItemStateEvent(OpenhabEvent, HABApp.core.events.ValueUpdateEvent):
    def __init__(self, name: str = '', value: typing.Any = None):
        super().__init__()

        # smarthome/items/NAME/state
        self.name: str = name
        self.value: typing.Any = value

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # smarthome/items/NAME/state
        return cls(topic[NAME_START:-6], map_openhab_values(payload['type'], payload['value']))

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}>'


class ItemStateChangedEvent(OpenhabEvent, HABApp.core.events.ValueChangeEvent):
    def __init__(self, name: str = '', value: typing.Any = None, old_value: typing.Any = None):
        super().__init__()

        self.name: str = name
        self.value = value
        self.old_value = old_value

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # smarthome/items/Ping/statechanged
        return cls(
            topic[NAME_START:-13],
            map_openhab_values(payload['type'], payload['value']),
            map_openhab_values(payload['oldType'], payload['oldValue'])
        )

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}, old_value: {self.old_value}>'


class ItemCommandEvent(OpenhabEvent):
    def __init__(self, name: str = '', value: typing.Any = None):
        super().__init__()

        self.name: str = name
        self.value = value

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # smarthome/items/NAME/command
        return cls(topic[NAME_START:-8], map_openhab_values(payload['type'], payload['value']))

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}>'


class ItemAddedEvent(OpenhabEvent):
    def __init__(self, name: str = '', type: str = ''):
        super().__init__()

        self.name: str = name
        self.type: str = type

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # {'topic': 'smarthome/items/NAME/added'
        # 'payload': '{"type":"Contact","name":"Test","tags":[],"groupNames":[]}'
        # 'type': 'ItemAddedEvent'}
        return cls(payload['name'], payload['type'])

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, type: {self.type}>'


class ItemUpdatedEvent(OpenhabEvent):
    def __init__(self, name: str = '', type: str = ''):
        super().__init__()

        self.name: str = name
        self.type: str = type

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # smarthome/items/NAME/updated
        # 'payload': '[{"type":"Switch","name":"Test","tags":[],"groupNames":[]},
        #              {"type":"Contact","name":"Test","tags":[],"groupNames":[]}]',
        # 'type': 'ItemUpdatedEvent'
        return cls(topic[NAME_START:-8], payload[0]['type'])

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, type: {self.type}>'


class ItemRemovedEvent(OpenhabEvent):
    def __init__(self, name: str = ''):
        super().__init__()

        self.name = name

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # smarthome/items/Test/removed
        return cls(topic[NAME_START:-8])

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}>'


class ItemStatePredictedEvent(OpenhabEvent):
    def __init__(self, name: str = '', value: typing.Any = None):
        super().__init__()

        # smarthome/items/NAME/state
        self.name: str = name
        self.value: typing.Any = value

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # 'smarthome/items/NAME/statepredicted'
        return cls(topic[NAME_START:-15], map_openhab_values(payload['predictedType'], payload['predictedValue']))

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}>'


class GroupItemStateChangedEvent(OpenhabEvent):
    def __init__(self, name: str = '', item: str = '', value: typing.Any = None, old_value: typing.Any = None):
        super().__init__()

        self.name: str = name
        self.item: str = item

        self.value = value
        self.old_value = old_value

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # 'smarthome/items/TestGroupAVG/TestNumber1/statechanged'
        parts = topic.split('/')

        return cls(
            parts[2], parts[3],
            map_openhab_values(payload['type'], payload['value']),
            map_openhab_values(payload['oldType'], payload['oldValue'])
        )

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}, old_value: {self.old_value}>'



import itertools
import logging
import traceback


class ShutdownHelper:
    def __init__(self, name=''):
        self.__cb_funcs = []
        self.__cb_last = []

        self.requested = False

    def register_func(self, func, last=False):
        assert callable(func)
        if last:
            self.__cb_last.append(func)
        else:
            self.__cb_funcs.append(func)

    def request_shutdown(self):
        "Request execution of all functions"

        log = logging.getLogger('HABApp.Shutdown')
        log.debug('Requested shutdown')

        self.requested = True

        for func in itertools.chain(self.__cb_funcs, self.__cb_last):
            try:
                log.debug(f'Calling {func.__name__}')
                func()
                log.debug(f'{func.__name__} done!')
            except Exception as ex:
                log.error(ex)
                tb = traceback.format_exc().splitlines()
                for line in tb:
                    log.error(line)

        log.debug('Shutdown complete')
        return None



import asyncio
import concurrent.futures
import io
import logging
import time
from cProfile import Profile
from pstats import Stats
try:
    from pstats import SortKey   # type: ignore[attr-defined]
    STAT_SORT_KEY = SortKey.CUMULATIVE
except ImportError:
    STAT_SORT_KEY = 'cumulative', 'cumtime'

import HABApp


default_logger = logging.getLogger('HABApp.Worker')


class WrappedFunction:
    _WORKERS = concurrent.futures.ThreadPoolExecutor(10, 'HabApp_')
    _EVENT_LOOP = None

    def __init__(self, func, logger=None, warn_too_long=True, name=None):
        assert callable(func)
        self._func = func

        # name of the function
        self.name = self._func.__name__ if not name else name

        self.is_async = asyncio.iscoroutinefunction(self._func)

        self.__time_submitted = 0.0

        # Allow custom logger
        self.log = default_logger
        if logger:
            self.log = logger

        self.__warn_too_long = warn_too_long

    def run(self, *args, **kwargs):

        if self.is_async:
            # schedule run async, we need to pass the event loop because we can create an async WrappedFunction
            # from a worker thread (if we have a mixture between async and non-async)!
            asyncio.run_coroutine_threadsafe(self.async_run(*args, **kwargs), loop=WrappedFunction._EVENT_LOOP)
        else:
            self.__time_submitted = time.time()
            WrappedFunction._WORKERS.submit(self.__run, *args, **kwargs)

    def __format_traceback(self, e: Exception, *args, **kwargs):

        lines = HABApp.core.wrapper.format_exception(e)

        # Log Exception
        self.log.error(f'Error in {self.name}: {e}')
        for line in lines:
            self.log.error(line)

        # create HABApp event, but only if we are not currently processing one
        if not args or not isinstance(args[0], HABApp.core.events.habapp_events.HABAppError):
            HABApp.core.EventBus.post_event(
                HABApp.core.const.topics.ERRORS, HABApp.core.events.habapp_events.HABAppError(
                    func_name=self.name, exception=e, traceback='\n'.join(lines)
                )
            )

    async def async_run(self, *args, **kwargs):
        try:
            await self._func(*args, **kwargs)
        except Exception as e:
            self.__format_traceback(e, *args, **kwargs)
        return None

    def __run(self, *args, **kwargs):

        __start = time.time()

        # notify if we don't process quickly
        if __start - self.__time_submitted > 0.05:
            self.log.warning(f'Starting of {self.name} took too long: {__start - self.__time_submitted:.2f}s. '
                             f'Maybe there are not enough threads?')

        # start profiler
        pr = Profile()
        pr.enable()

        # Execute the function
        try:
            self._func(*args, **kwargs)
        except Exception as e:
            self.__format_traceback(e, *args, **kwargs)

        # disable profiler
        pr.disable()

        # log warning if execution takes too long
        __dur = time.time() - __start
        if self.__warn_too_long and __dur > 0.8:
            self.log.warning(f'Execution of {self.name} took too long: {__dur:.2f}s')

            s = io.StringIO()
            ps = Stats(pr, stream=s).sort_stats(STAT_SORT_KEY)
            ps.print_stats(0.1)  # limit to output to 10% of the lines

            for line in s.getvalue().splitlines()[4:]:    # skip the amount of calls and "Ordered by:"
                if line:
                    self.log.warning(line)



import re
import typing

import bidict

from HABApp.core.logger import log_error
from HABApp.openhab.connection_handler.func_async import async_set_thing_cfg
from ._log import log_cfg as log


def ensure_same_types(a, b, key: str):
    t_a = type(a)
    t_b = type(b)
    if t_a is t_b:
        return None

    _a = str(t_a)
    _a = _a[7:-1] if _a.startswith('<class ') and _a[-1] == '>' else _a

    _b = str(t_b)
    _b = _b[7:-1] if _b.startswith('<class ') and _b[-1] == '>' else _b
    raise ValueError(f"Datatype of parameter '{key}' must be {_a} but is {_b}!")


class ThingConfigChanger:
    zw_param = re.compile(r'config_(?P<param>\d+)_(?P<width>\d+)(?P<bitmask>_\w+)?')
    zw_group = re.compile(r'group_(\d+)')

    @classmethod
    def from_dict(cls, uid: str, _in: dict) -> 'ThingConfigChanger':
        c = cls(uid)
        c.org = _in
        for k in _in:
            # Z-Wave Params -> 0
            m = ThingConfigChanger.zw_param.fullmatch(k)
            if m:
                # check if the entry without bitmask is also in the cfg dict because then we use this entry
                # and split up the bitmask entries accordingly
                # 154 -> config_154_4
                # 154_000000FF -> config_154_4_000000FF
                if m.group(3) is not None:
                    if f'config_{m.group(1)}_{m.group(2)}' in _in:
                        c.alias[f'{m.group(1)}{m.group(3)}'] = k
                        continue

                c.alias[int(m.group(1))] = k
                continue

            # Z-Wave Groups to Group1
            m = ThingConfigChanger.zw_group.fullmatch(k)
            if m:
                c.alias[f'Group{m.group(1)}'] = k
                continue
        return c

    def __init__(self, uid: str):
        self.uid: str = uid
        self.alias = bidict.bidict()
        self.org: typing.Dict[str, typing.Any] = {}
        self.new: typing.Dict[str, typing.Any] = {}

    def __getitem__(self, key):
        return self.org[self.alias.get(key, key)]

    def __setitem__(self, o_key, item):
        key = self.alias.get(o_key, o_key)
        if key not in self.org:
            raise KeyError(f'Parameter "{o_key}" does not exist for {self.uid}!')

        org = self.org[key]
        ensure_same_types(item, org, o_key)

        if item == org:
            return None
        self.new[key] = item

    def __contains__(self, key):
        return self.alias.get(key, key) in self.org

    def get(self, key, default=None):
        try:
            return self.org[self.alias.get(key, key)]
        except KeyError:
            return default

    def keys(self):
        return (self.alias.inverse.get(k, k) for k in self.org.keys())

    def values(self):
        return self.org.values()

    def get_dict(self, filter=False, new=False):
        r = {}
        for k, v in (self.org if not new else self.new).items():
            if filter and (k.startswith('action_') or k in ('node_id', 'wakeup_node')):
                continue
            k = self.alias.inverse.get(k, k)
            r[k] = v
        return r

    async def update_thing_cfg(self):
        if not self.new:
            log.debug(f'Config of {self.uid} is already correct!')
            return None

        try:
            # we write only the changed configuration
            ret = await async_set_thing_cfg(self.uid, cfg=self.new)
            if ret is None:
                return None
        except Exception as e:
            log_error(log, f'Could not set new config for {self.uid}: {e}')
            return None

        log.info(f'Config of {self.uid} successfully updated!')



import HABApp
from HABApp.core.events import AllEvents
from . import WrappedFunction
from typing import Optional, Any


class EventBusListener:
    def __init__(self, topic, callback, event_type=AllEvents,
                 prop_name1: Optional[str] = None, prop_value1: Optional[Any] = None,
                 prop_name2: Optional[str] = None, prop_value2: Optional[Any] = None,
                 ):
        assert isinstance(topic, str), type(topic)
        assert isinstance(callback, WrappedFunction)
        assert prop_name1 is None or isinstance(prop_name1, str), prop_name1
        assert prop_name2 is None or isinstance(prop_name2, str), prop_name2

        self.topic: str = topic
        self.func: WrappedFunction = callback

        self.event_filter = event_type

        # Property filters
        self.prop_name1 = prop_name1
        self.prop_value1 = prop_value1
        self.prop_name2 = prop_name2
        self.prop_value2 = prop_value2

        self.__is_all: bool = self.event_filter is AllEvents
        self.__is_single: bool = not isinstance(self.event_filter, (list, tuple, set))

    def notify_listeners(self, event):
        # We run always
        if self.__is_all:
            self.func.run(event)
            return None

        # single filter
        if self.__is_single:
            if isinstance(event, self.event_filter):
                # If we have property filters wie only trigger when value is set accordingly
                if self.prop_name1 is not None:
                    if getattr(event, self.prop_name1, None) != self.prop_value1:
                        return None
                if self.prop_name2 is not None:
                    if getattr(event, self.prop_name2, None) != self.prop_value2:
                        return None

                self.func.run(event)
            return None

        # Make it possible to specify multiple classes
        for cls in self.event_filter:
            if isinstance(event, cls):
                # If we have property filters wie only trigger when value is set accordingly
                if self.prop_name1 is not None:
                    if getattr(event, self.prop_name1, None) != self.prop_value1:
                        return None
                if self.prop_name2 is not None:
                    if getattr(event, self.prop_name2, None) != self.prop_value2:
                        return None

                self.func.run(event)
                return None

    def cancel(self):
        """Stop listening on the event bus"""
        HABApp.core.EventBus.remove_listener(self)

    def desc(self):
        # return description
        _type = str(self.event_filter)
        if _type.startswith("<class '"):
            _type = _type[8:-2].split('.')[-1]

        _val = ''
        if self.prop_name1 is not None:
            _val += f', {self.prop_name1}=={self.prop_value1}'
        if self.prop_name2 is not None:
            _val += f', {self.prop_name2}=={self.prop_value2}'

        return f'"{self.topic}" (type {_type}{_val})'



import datetime
import logging
import typing
import warnings
from threading import Lock

import HABApp
from HABApp.core.items import Item
from HABApp.rule import get_parent_rule
from .mode_base import BaseMode

LOCK = Lock()


class MultiModeItem(Item):
    """Prioritizer :class:`~HABApp.core.items.Item`

    :ivar logger: Assign a logger to get log messages about the different modes
    """

    @classmethod
    def get_create_item(cls, name: str, logger=None, initial_value=None):
        # added 20.04.2020, van be removed in some time
        if logger is not None:
            warnings.warn("'logger' is deprecated, set logger on the mode instead!", DeprecationWarning, 2)

        return super().get_create_item(name, initial_value)

    def __init__(self, name: str, initial_value=None):
        super().__init__(name=name, initial_value=initial_value)

        self.__values_by_prio: typing.Dict[int, BaseMode] = {}
        self.__values_by_name: typing.Dict[str, BaseMode] = {}

    def __remove_mode(self, name: str) -> bool:

        # Check if the mode exists
        found = self.__values_by_name.pop(name, None)
        if found is None:
            return False

        # Remove parent mapping, so we at least crash if someone attempts to do something with the mode
        found.parent = None

        # remove value entry, too
        for i, f in self.__values_by_prio.items():
            if f is found:
                self.__values_by_prio.pop(i)
                return True
        raise RuntimeError()

    def __sort_modes(self):
        # make the lower priority known to the mode
        low = None
        for _, child in sorted(self.__values_by_prio.items()):  # type: int, BaseMode
            child._set_mode_lower_prio(low)
            low = child
        return None

    def remove_mode(self, name: str) -> bool:
        """Remove mode if it exists

        :param name: name of the mode (case insensitive)
        :return: True if something was removed, False if nothign was found
        """
        assert isinstance(name, str), type(name)

        with LOCK:
            found = self.__remove_mode(name.lower())
            self.__sort_modes()
            return found

    def add_mode(self, priority: int, mode: BaseMode) -> 'MultiModeItem':
        """Add a new mode to the item, if it already exists it will be overwritten

        :param priority: priority of the mode
        :param mode: instance of the MultiMode class
        """
        assert isinstance(priority, int), type(priority)
        assert isinstance(mode, BaseMode), type(mode)

        name = mode.name.lower()

        with LOCK:
            # remove old mode
            self.__remove_mode(name)

            # add new mode
            mode.parent = self
            self.__values_by_prio[priority] = mode
            self.__values_by_name[name] = mode

            # resort
            self.__sort_modes()

        try:
            get_parent_rule().register_cancel_obj(mode)
        except RuntimeError:
            HABApp.core.logger.log_warning(
                logger=logging.getLogger('HABApp'), text='Parent rule not found! '
                f'Automatic unloading of the {self.__class__.__name__} {self.name} will not work!'
            )
        return self

    def all_modes(self) -> typing.List[typing.Tuple[int, BaseMode]]:
        """Returns a sorted list containing tuples with the priority and the mode

        :return: List with priorities and modes
        """
        return sorted(self.__values_by_prio.items())

    def get_mode(self, name: str) -> BaseMode:
        """Returns a created mode

        :param name: name of the mode (case insensitive)
        :return: The requested MultiModeValue
        """
        try:
            return self.__values_by_name[name.lower()]
        except KeyError:
            raise KeyError(f'Unknown mode "{name}"! Available: {", ".join(self.__values_by_name.keys())}') from None

    def calculate_value(self) -> typing.Any:
        """Recalculate the output value and post the state to the event bus (if it is not None)

        :return: new value
        """

        # recalculate value
        new_value = None
        with LOCK:
            for priority, child in sorted(self.__values_by_prio.items()):
                new_value = child.calculate_value(new_value)

        if new_value is not None:
            self.post_value(new_value)
        return new_value

    def create_mode(
            self, name: str, priority: int, initial_value: typing.Optional[typing.Any] = None,
            auto_disable_after: typing.Optional[datetime.timedelta] = None,
            auto_disable_func: typing.Optional[typing.Callable[[typing.Any, typing.Any], bool]] = None,
            calc_value_func: typing.Optional[typing.Callable[[typing.Any, typing.Any], typing.Any]] = None
    ):
        warnings.warn("'create_mode' is deprecated, create a mode and pass it to 'add_mode' instead!",
                      DeprecationWarning, 2)

        m = HABApp.util.multimode.ValueMode(name=name, initial_value=initial_value)
        m.auto_disable_after = auto_disable_after
        m.auto_disable_func = auto_disable_func
        m.calc_value_func = calc_value_func

        self.add_mode(priority, m)
        return m



import HABApp.mqtt.mqtt_interface
from HABApp.core.items import Item


class MqttItem(Item):

    def publish(self, payload, qos: int = None, retain: bool = None):
        """
        Publish the payload under the topic from the item.

        :param payload: MQTT Payload
        :param qos: QoS, can be ``0``, ``1`` or ``2``. If not specified value from configuration file will be used.
        :param retain: retain message. If not specified value from configuration file will be used.
        :return: 0 if successful
        """

        return HABApp.mqtt.mqtt_interface.MQTT_INTERFACE.publish(self.name, payload, qos=qos, retain=retain)



import random
import typing
import tzlocal
from datetime import datetime, timedelta, time

from pytz import utc

from HABApp.core import WrappedFunction

TYPING_DATE_TIME = typing.Union[None, datetime, timedelta, time]

local_tz = tzlocal.get_localzone()


def replace_time(dt_obj: datetime, t_obj: typing.Union[datetime, time]) -> datetime:
    return dt_obj.replace(hour=t_obj.hour, minute=t_obj.minute, second=t_obj.second, microsecond=0)


class ScheduledCallbackBase:

    def __init__(self, callback: WrappedFunction, *args, **kwargs):

        self._callback: WrappedFunction = callback
        self._args = args
        self._kwargs = kwargs

        # boundaries
        self._earliest: typing.Optional[time] = None
        self._latest: typing.Optional[time] = None
        self._offset: typing.Optional[timedelta] = None
        self._jitter: typing.Optional[int] = None
        self._boundary_func: typing.Optional[typing.Callable[[datetime], datetime]] = None

        # times when we run
        self._next_base: datetime
        self._next_call: datetime

        # properties
        self.is_due = False
        self.is_finished = False
        self.run_counter = 0

    def _set_next_base(self, next_time: TYPING_DATE_TIME) -> 'ScheduledCallbackBase':
        # initializer method for self._next_base

        __now = datetime.now()
        if next_time is None:
            # If we don't specify a datetime we start it now
            base_time = __now
        elif isinstance(next_time, timedelta):
            # if it is a timedelta add it to now to easily speciy points in the future
            base_time = __now + next_time
        elif isinstance(next_time, time):
            # if it is a time object it specifies a time of day.
            base_time = __now.replace(hour=next_time.hour, minute=next_time.minute, second=next_time.second)
            if base_time < __now:
                base_time += timedelta(days=1)
        else:
            base_time = next_time

        assert isinstance(base_time, datetime), type(base_time)

        # convert to utc
        base_time = local_tz.localize(base_time)
        self._next_base = base_time.astimezone(utc)
        return self

    def _calculate_next_call(self):
        """Get next calc time"""
        raise NotImplementedError()

    def earliest(self, time_obj: typing.Optional[time]) -> 'ScheduledCallbackBase':
        """Set earliest boundary as time of day. ``None`` will disable boundary.

        :param time_obj: time obj, scheduler will not run earlier
        """
        assert isinstance(time_obj, time) or time_obj is None, type(time_obj)
        changed = self._earliest != time_obj
        self._earliest = time_obj
        if changed:
            self._update_run_time()
        return self

    def latest(self, time_obj: typing.Optional[time]) -> 'ScheduledCallbackBase':
        """Set earliest boundary as time of day. ``None`` will disable boundary.

        :param time_obj: time obj, scheduler will not run later
        """
        assert isinstance(time_obj, time) or time_obj is None, type(time_obj)
        changed = self._latest != time_obj
        self._latest = time_obj
        if changed:
            self._update_run_time()
        return self

    def offset(self, timedelta_obj: typing.Optional[timedelta]) -> 'ScheduledCallbackBase':
        """Set a constant offset to the calculation of the next run. ``None`` will disable the offset.

        :param timedelta_obj: constant offset
        """
        assert isinstance(timedelta_obj, timedelta) or timedelta_obj is None, type(timedelta_obj)
        changed = self._offset != timedelta_obj
        self._offset = timedelta_obj
        if changed:
            self._update_run_time()
        return self

    def jitter(self, secs: typing.Optional[int]) -> 'ScheduledCallbackBase':
        """Add a random jitter per call in the intervall [(-1) * secs ... secs] to the next run.
         ``None`` will disable jitter.

        :param secs: jitter in secs
        """
        assert isinstance(secs, int) or secs is None, type(secs)
        changed = self._jitter != secs
        self._jitter = secs
        if changed:
            self._update_run_time()
        return self

    def boundary_func(self, func: typing.Optional[typing.Callable[[datetime], datetime]]):
        """Add a function which will be called when the datetime changes. Use this to implement custom boundaries.
         Use ``None`` to disable the boundary function.

        :param func: Function which returns a datetime obj, arg is a datetime with the next call time
        """
        changed = self._boundary_func != func
        self._boundary_func = func
        if changed:
            self._update_run_time()
        return self

    def _update_run_time(self) -> 'ScheduledCallbackBase':
        # Starting point is always the next call
        self._next_call = self._next_base

        # custom boundaries first
        if self._boundary_func is not None:
            self._next_call = self._boundary_func(self._next_call.astimezone(local_tz)).astimezone(utc)

        if self._offset is not None:
            self._next_call += self._offset  # offset doesn't have to be localized

        if self._jitter is not None:
            self._next_call += timedelta(seconds=random.randint(-1 * self._jitter, self._jitter))

        if self._earliest is not None:
            earliest = replace_time(self._next_call.astimezone(local_tz), self._earliest)
            earliest = earliest.astimezone(utc)
            if self._next_call < earliest:
                self._next_call = earliest

        if self._latest is not None:
            latest = replace_time(self._next_call.astimezone(local_tz), self._latest)
            latest = latest.astimezone(utc)
            if self._next_call > latest:
                self._next_call = latest

        return self

    def get_next_call(self):
        """Return the next execution timestamp"""
        return self._next_call.astimezone(local_tz).replace(tzinfo=None)

    def check_due(self, now: datetime):
        """Check whether the callback is due for execution

        :param now:
        :return:
        """

        self.is_due = True if self._next_call <= now else False
        if self.is_finished:
            self.is_due = False

        return self.is_due

    def execute(self) -> bool:
        """Try to execute callback. If the callback is not due yet or execution has already finished nothing will happen

        :return: True if callback has been executed else False
        """
        if not self.is_due or self.is_finished:
            return False

        self.run_counter += 1
        self._calculate_next_call()
        self._callback.run(*self._args, **self._kwargs)
        return True

    def cancel(self):
        """Cancel execution
        """
        self.is_finished = True



import asyncio
import logging

import HABApp
from HABApp.openhab.definitions.helpers.log_table import Table
from ._plugin import OnConnectPlugin
from ..interface_async import async_get_things


class ThingOverview(OnConnectPlugin):


    def __init__(self):
        super().__init__()
        self.run = False

    @HABApp.core.wrapper.log_exception
    async def on_connect_function(self):
        # don't run this after the connect, let the rules load etc.
        await asyncio.sleep(60)

        # show this overview only once!
        if self.run:
            return None
        self.run = True

        thing_data = await async_get_things()

        thing_table = Table('Things')
        thing_stat = thing_table.add_column('Status', '^')
        thing_label = thing_table.add_column('Label')
        thing_location = thing_table.add_column('Location')
        thing_type = thing_table.add_column('Thing type')
        thing_uid = thing_table.add_column('Thing UID')

        zw_table = Table(heading='Z-Wave Things')
        zw_node = zw_table.add_column('Node', '^')
        zw_stat = zw_table.add_column('Status', '^')
        zw_label = zw_table.add_column('Label')
        zw_location = zw_table.add_column('Location')
        zw_model = zw_table.add_column('Model', '^')
        zw_fw = zw_table.add_column('Firmware', '^')
        zw_type = zw_table.add_column('Thing type')
        zw_uid = zw_table.add_column('Thing UID')
        zw_l_channels = zw_table.add_column('Linked channel types')
        zw_u_channels = zw_table.add_column('Unlinked channel types')

        for node in thing_data:
            uid = node['UID']
            type_uid = node['thingTypeUID']

            is_zw = type_uid.startswith('zwave:')

            col_uid, col_stat, col_label, col_location, col_type = \
                (thing_uid, thing_stat, thing_label, thing_location, thing_type) if not is_zw else \
                (zw_uid, zw_stat, zw_label, zw_location, zw_type)

            col_uid.add(uid)
            col_type.add(type_uid)
            col_label.add(node.get('label', ''))
            col_stat.add(node['statusInfo']['status'])
            col_location.add(node.get('location', ''))

            if not is_zw:
                continue

            # optional properties which can be set
            props = node['properties']
            channels = node.get('channels', [])

            # Node-ID, e.g. 5
            node_id = props.get('zwave_nodeid')
            zw_node.add(int(node_id) if node_id is not None else '')

            zw_model.add(props.get('modelId', ''))
            zw_fw.add(props.get('zwave_version', ''))

            zw_l_channels.add(
                ', '.join(map(lambda x: x.get('channelTypeUID', ''), filter(lambda x: x.get('linkedItems'), channels)))
            )
            zw_u_channels.add(', '.join(
                map(lambda x: x.get('channelTypeUID', ''), filter(lambda x: not x.get('linkedItems'), channels)))
            )

        log = logging.getLogger('HABApp.openhab.things')
        for line in thing_table.get_lines(sort_columns=[thing_uid]):
            log.info(line)

        log = logging.getLogger('HABApp.openhab.zwave')
        for line in zw_table.get_lines(sort_columns=[zw_type, 'Node']):
            log.info(line)


PLUGIN_THING_OVERVIEW = ThingOverview.create_plugin()



import logging
import sys
from pathlib import Path

from EasyCo import ConfigFile, PathContainer, ConfigEntry

from ._conf_location import Location
from ._conf_mqtt import Mqtt
from ._conf_openhab import Openhab

log = logging.getLogger('HABApp.Config')


class Directories(PathContainer):
    logging: Path = ConfigEntry(Path('log'), description='Folder where the logs will be written to')
    rules: Path = ConfigEntry(Path('rules'), description='Folder from which the rule files will be loaded')
    param: Path = ConfigEntry(Path('param'), description='Folder from which the parameter files will be loaded')
    config: Path = ConfigEntry(Path('config'), description='Folder from which configuration files will be loaded')
    lib: Path = ConfigEntry(Path('lib'), description='Folder where additional libraries can be placed')

    def on_all_values_set(self):
        try:
            # create folder structure if it does not exist
            if not self.rules.is_dir():
                self.rules.mkdir()
            if not self.logging.is_dir():
                self.logging.mkdir()
            if not self.config.is_dir():
                log.info(f'Manual thing configuration disabled! Folder {self.config} does not exist!')

            # add path for libraries
            if self.lib.is_dir():
                lib_path = str(self.lib)
                if lib_path not in sys.path:
                    sys.path.insert(0, lib_path)
                    log.debug(f'Added library folder "{lib_path}" to system path')
        except Exception as e:
            log.error(e)
            print(e)


class HABAppConfig(ConfigFile):
    location = Location()
    directories = Directories()
    mqtt = Mqtt()
    openhab = Openhab()


CONFIG: HABAppConfig = HABAppConfig()



# no external dependencies
import HABApp.openhab.exceptions
import HABApp.openhab.events

import HABApp.openhab.interface_async
import HABApp.openhab.interface

# items use the interface for the convenience functions
import HABApp.openhab.items



from pathlib import Path

from watchdog.events import FileSystemEvent, FileSystemEventHandler


class BaseWatcher(FileSystemEventHandler):
    def __init__(self, folder: Path, file_ending: str, watch_subfolders: bool = False):
        assert isinstance(folder, Path), type(folder)
        assert isinstance(file_ending, str), type(file_ending)
        assert watch_subfolders is True or watch_subfolders is False

        self.folder: Path = folder
        self.file_ending: str = file_ending
        self.watch_subfolders: bool = watch_subfolders

    def dispatch(self, event: FileSystemEvent):
        # we don't process directory events
        if event.is_directory:
            return None

        src = event.src_path
        if src.endswith(self.file_ending):
            self.file_changed(src)

        # moved events have a dst, so we process it, too
        if hasattr(event, 'dest_path'):
            dst = event.dest_path
            if dst.endswith(self.file_ending):
                self.file_changed(dst)
        return None

    def file_changed(self, dst: str):
        raise NotImplementedError()



import argparse
import asyncio
import logging
import os
import signal
import sys
import time
import traceback
import typing
from pathlib import Path

import HABApp


def find_config_folder(arg_config_path: typing.Optional[Path]) -> Path:

    if arg_config_path is None:
        # Nothing is specified, we try to find the config automatically
        check_path = []
        try:
            working_dir = Path(os.getcwd())
            check_path.append( working_dir / 'HABApp')
            check_path.append( working_dir.with_name('HABApp'))
            check_path.append( working_dir.parent.with_name('HABApp'))
        except ValueError:
            # the ValueError gets raised if the working_dir or its parent is empty (e.g. c:\ or /)
            pass

        check_path.append(Path.home() / 'HABApp')   # User home

        # if we run in a venv check the venv, too
        v_env = os.environ.get('VIRTUAL_ENV', '')
        if v_env:
            check_path.append(Path(v_env) / 'HABApp')  # Virtual env dir
    else:
        # in case the user specifies the config.yml we automatically switch to the parent folder
        if arg_config_path.name.lower() == 'config.yml' and arg_config_path.is_file():
            arg_config_path = arg_config_path.parent

        # Override automatic config detection if something is specified through command line
        check_path = [arg_config_path]

    for config_folder in check_path:
        config_folder = config_folder.resolve()
        if not config_folder.is_dir():
            continue

        config_file = config_folder / 'config.yml'
        if config_file.is_file():
            return config_folder

    # we have specified a folder, but the config does not exist so we will create it
    if arg_config_path is not None and arg_config_path.is_dir():
        return arg_config_path

    # we have nothing found and nothing specified -> exit
    print('Config file "config.yml" not found!')
    print('Checked folders:\n - ' + '\n - '.join(str(k) for k in check_path))
    print('Please create file or specify a folder with the "-c" arg switch.')
    sys.exit(1)


def get_command_line_args(args=None):
    parser = argparse.ArgumentParser(description='Start HABApp')
    parser.add_argument(
        '-c',
        '--config',
        help='Path to configuration folder (where the config.yml is located)',
        default=None
    )
    parser.add_argument(
        '-s',
        '--sleep',
        help='Sleep time in seconds before starting HABApp',
        type=int,
        default=None
    )
    return parser.parse_args(args)


def main() -> typing.Union[int, str]:

    args = get_command_line_args()
    if args.sleep:
        args.sleep = max(0, args.sleep)
        print(f'Waiting {args.sleep:d} seconds before starting HABApp ...', end='')
        time.sleep(args.sleep)
        print(' done!')
    if args.config is not None:
        args.config = Path(args.config).resolve()

    log = logging.getLogger('HABApp')

    try:
        app = HABApp.runtime.Runtime()
        app.startup(config_folder=find_config_folder(args.config))

        def shutdown_handler(sig, frame):
            print('Shutting down ...')
            app.shutdown.request_shutdown()

        # register shutdown helper
        signal.signal(signal.SIGINT, shutdown_handler)
        signal.signal(signal.SIGTERM, shutdown_handler)

        # start workers
        try:
            HABApp.core.const.loop.run_until_complete(app.get_async())
        except asyncio.CancelledError:
            pass
    except HABApp.config.InvalidConfigException:
        pass
    except Exception as e:
        for line in traceback.format_exc().splitlines():
            log.error(line)
            print(e)
        return str(e)
    finally:
        # Sleep to allow underlying connections of aiohttp to close
        # https://aiohttp.readthedocs.io/en/stable/client_advanced.html#graceful-shutdown
        HABApp.core.const.loop.run_until_complete(asyncio.sleep(1))
        HABApp.core.const.loop.close()
    return 0


if __name__ == "__main__":
    sys.exit(main())



import datetime
import logging
import typing

from HABApp.core.items.base_item import BaseItem
from HABApp.core.wrapper import process_exception
from HABApp.openhab.definitions.values import QuantityValue, RawValue
from HABApp.openhab.items import ColorItem, ContactItem, DatetimeItem, DimmerItem, GroupItem, ImageItem, LocationItem, \
    NumberItem, PlayerItem, RollershutterItem, StringItem, SwitchItem

log = logging.getLogger('HABApp.openhab')


def map_item(name, openhab_type: str, openhab_value: str) -> typing.Optional[BaseItem]:
    try:
        assert isinstance(openhab_type, str), type(openhab_type)
        assert isinstance(openhab_value, str), type(openhab_value)

        value: typing.Optional[str] = openhab_value
        if openhab_value == 'NULL' or openhab_value == 'UNDEF':
            value = None

        # Quantity types are like this: Number:Temperature and have a unit set: "12.3 °C".
        # We have to remove the dimension from the type and remove the unit from the value
        if ':' in openhab_type:
            openhab_type, dimension = openhab_type.split(':')
            # if the item is not initialized its None and has no dimension
            if value is not None:
                value, _ = QuantityValue.split_unit(value)

        # Specific classes
        if openhab_type == "Switch":
            return SwitchItem(name, value)

        if openhab_type == "String":
            return StringItem(name, value)

        if openhab_type == "Contact":
            return ContactItem(name, value)

        if openhab_type == "Rollershutter":
            if value is None:
                return RollershutterItem(name, value)
            return RollershutterItem(name, float(value))

        if openhab_type == "Dimmer":
            if value is None:
                return DimmerItem(name, value)
            return DimmerItem(name, float(value))

        if openhab_type == "Number":
            if value is None:
                return NumberItem(name, value)

            # Number items can be int or float
            try:
                return NumberItem(name, int(value))
            except ValueError:
                return NumberItem(name, float(value))

        if openhab_type == "DateTime":
            if value is None:
                return DatetimeItem(name, value)
            dt = datetime.datetime.strptime(value.replace('+', '000+'), '%Y-%m-%dT%H:%M:%S.%f%z')
            # all datetimes from openhab have a timezone set so we can't easily compare them
            # --> TypeError: can't compare offset-naive and offset-aware datetimes
            dt = dt.astimezone(tz=None)   # Changes datetime object so it uses system timezone
            dt = dt.replace(tzinfo=None)  # Removes timezone awareness
            return DatetimeItem(name, dt)

        if openhab_type == "Color":
            if value is None:
                return ColorItem(name)
            return ColorItem(name, *(float(k) for k in value.split(',')))

        if openhab_type == "Image":
            img = ImageItem(name)
            if value is None:
                return img
            img.set_value(RawValue(value))
            return img

        if openhab_type == "Group":
            return GroupItem(name, value)

        if openhab_type == "Location":
            return LocationItem(name, value)

        if openhab_type == "Player":
            return PlayerItem(name, value)

        raise ValueError(f'Unknown Openhab type: {openhab_type} for {name}')

    except Exception as e:
        process_exception('map_items', e, logger=log)
        return None



import logging
import typing
from pathlib import Path

import HABApp
from HABApp.core.const.topics import FILES as T_FILES
from HABApp.core.events.habapp_events import RequestFileLoadEvent, RequestFileUnloadEvent
from .file_props import FileProperties, get_props

log = logging.getLogger('HABApp.files')


class CircularReferenceError(Exception):
    pass


class HABAppFile:

    @classmethod
    def from_path(cls, name: str, path: Path) -> 'HABAppFile':
        with path.open('r', encoding='utf-8') as f:
            txt = f.read(10 * 1024)
        return cls(name, path, get_props(txt))

    def __init__(self, name: str, path: Path, properties: FileProperties):
        self.name: str = name
        self.path: Path = path
        self.properties: FileProperties = properties

        # file checks
        self.is_checked = False
        self.is_valid = False

        # file loaded
        self.is_loaded = False
        self.is_failed = False

    def _check_refs(self, stack, prop: str):
        c: typing.List[str] = getattr(self.properties, prop)
        for f in c:
            _stack = stack + (f, )
            if f in stack:
                log.error(f'Circular reference: {" -> ".join(_stack)}')
                raise CircularReferenceError(" -> ".join(_stack))

            next_file = ALL.get(f)
            if next_file is not None:
                next_file._check_refs(_stack, prop)

    def check_properties(self):
        self.is_checked = True

        # check dependencies
        mis = set(filter(lambda x: x not in ALL, self.properties.depends_on))
        if mis:
            one = len(mis) == 1
            msg = f'File {self.path} depends on file{"" if one else "s"} that ' \
                  f'do{"es" if one else ""}n\'t exist: {", ".join(sorted(mis))}'
            log.error(msg)
            raise FileNotFoundError(msg)

        # check reload
        mis = set(filter(lambda x: x not in ALL, self.properties.reloads_on))
        if mis:
            one = len(mis) == 1
            log.warning(f'File {self.path} reloads on file{"" if one else "s"} that '
                        f'do{"es" if one else ""}n\'t exist: {", ".join(sorted(mis))}')

        # check for circular references
        self._check_refs((self.name, ), 'depends_on')
        self._check_refs((self.name, ), 'reloads_on')

        self.is_valid = True

    def can_be_loaded(self) -> bool:
        if not self.is_valid:
            return False

        for name in self.properties.depends_on:
            f = ALL.get(name, None)
            if f is None:
                return False

            if not f.is_loaded:
                return False
        return True

    def load(self):
        self.is_loaded = False

        HABApp.core.EventBus.post_event(
            T_FILES, RequestFileLoadEvent(self.name)
        )

    def unload(self):
        HABApp.core.EventBus.post_event(
            T_FILES, RequestFileUnloadEvent(self.name)
        )

    def load_ok(self):
        self.is_loaded = True

    def load_failed(self):
        self.is_failed = True


from .all import ALL    # noqa F401



import typing

import paho.mqtt.client as mqtt

from .mqtt_connection import MqttConnection, log
from ..config import Mqtt as MqttConfig
from ..core.const.json import dump_json


class MqttInterface:

    def __init__(self, connection: MqttConnection, config: MqttConfig):
        assert isinstance(connection, MqttConnection)
        assert isinstance(config, MqttConfig)

        self.__connection: MqttConnection = connection
        self.__config: MqttConfig = config

    def __is_connected(self) -> bool:
        if self.__connection.connected:
            return True

        raise ConnectionError('Mqtt client not connected')


    def publish(self, topic: str, payload: typing.Any, qos: int = None, retain: bool = None) -> int:
        """
        Publish a value under a certain topic.
        If qos and/or retain is not set the value from the configuration file will be used.

        :param topic: MQTT topic
        :param payload: MQTT Payload
        :param qos: QoS, can be 0, 1 or 2. If not specified value from configuration file will be used.
        :param retain: retain message. If not specified value from configuration file will be used.
        :return: 0 if successful
        """

        assert isinstance(topic, str), type(topic)
        assert isinstance(qos, int) or qos is None, type(qos)
        assert isinstance(retain, bool) or retain is None, type(retain)

        if not self.__is_connected():
            return mqtt.MQTT_ERR_NO_CONN
        if self.__config.general.listen_only:
            return 100

        if qos is None:
            qos = self.__config.publish.qos
        if retain is None:
            retain = self.__config.publish.retain

        # convert these to string
        if isinstance(payload, (dict, list)):
            payload = dump_json(payload)

        info = self.__connection.client.publish(topic, payload, qos, retain)
        if info.rc != mqtt.MQTT_ERR_SUCCESS:
            log.error(f'Could not publish to "{topic}": {mqtt.error_string(info.rc)}')
        return info

    def subscribe(self, topic: str, qos: int = None) -> int:
        """
        Subscribe to a MQTT topic. Subscriptions will be active until next disconnect

        :param topic: MQTT topic to subscribe to
        :param qos: QoS, can be 0, 1 or 2.  If not specified value from configuration file will be used.
        :return: 0 if successful
        """

        assert isinstance(topic, str), type(topic)
        assert isinstance(qos, int) or qos is None, type(qos)

        if not self.__is_connected():
            return mqtt.MQTT_ERR_NO_CONN

        # If no qos is specified load it from config
        if qos is None:
            qos = self.__config.subscribe.qos

        res, mid = self.__connection.client.subscribe(topic, qos)
        if res != mqtt.MQTT_ERR_SUCCESS:
            log.error(f'Could not subscribe to "{topic}": {mqtt.error_string(res)}')
        return res

    def unsubscribe(self, topic: str) -> int:
        """
        Unsubscribe from a MQTT topic

        :param topic: MQTT topic
        :return: 0 if successful
        """

        assert isinstance(topic, str), type(topic)

        if not self.__is_connected():
            return mqtt.MQTT_ERR_NO_CONN

        result, mid = self.__connection.client.unsubscribe(topic)
        if result != mqtt.MQTT_ERR_SUCCESS:
            log.error(f'Could not unsubscribe from "{topic}": {mqtt.error_string(result)}')
        return result


MQTT_INTERFACE: MqttInterface


def get_mqtt_interface(connection=None, config=None) -> MqttInterface:
    global MQTT_INTERFACE
    if connection is None:
        return MQTT_INTERFACE

    MQTT_INTERFACE = MqttInterface(connection, config)
    return MQTT_INTERFACE



from . import const
from . import lib
from . import wrapper
from . import logger

from .wrappedfunction import WrappedFunction

from .event_bus_listener import EventBusListener

import HABApp.core.events
import HABApp.core.files
import HABApp.core.items

import HABApp.core.EventBus
import HABApp.core.Items



import threading
import time


class PeriodCounter:
    def __init__(self, period):
        assert isinstance(period, int)
        self.period = period

        # Thread save
        self.__lock = threading.Lock()
        # funcs which gets called when the counter changes
        self.__on_change = set()

        self.__timestamps = []

    def on_change(self, func, unregister=False):
        assert callable(func)
        if unregister:
            self.__on_change.remove(func)
        else:
            self.__on_change.add(func)

    def __clean_timestamps(self, add=False):
        now = time.time()
        min_ts = now - self.period
        self.__timestamps = [k for k in self.__timestamps if k >= min_ts]
        if add:
            self.__timestamps.append(now)

    def reset(self):
        with self.__lock:
            count_was = len(self.__timestamps)
            self.__timestamps = []
            count_new = len(self.__timestamps)

        if count_was != count_new:
            for func in self.__on_change:
                func()

    def increase(self) -> int:
        with self.__lock:
            count_was = len(self.__timestamps)
            self.__clean_timestamps(add=True)
            count_new = len(self.__timestamps)

        if count_was != count_new:
            for func in self.__on_change:
                func(count_new)

        return count_new

    def get_count(self) -> int:
        with self.__lock:
            count_was = len(self.__timestamps)
            self.__clean_timestamps(add=False)
            count_new = len(self.__timestamps)

        if count_was != count_new:
            for func in self.__on_change:
                func(count_new)

        return count_new



from . import watcher
from .file_name import path_from_name, name_from_path
from .file import HABAppFile
from .all import watch_folder, file_load_failed, file_load_ok
from .event_listener import add_event_bus_listener



from datetime import datetime

from pytz import utc

from HABApp.core.items.base_item import BaseItem
from ..events import ThingStatusInfoEvent


class Thing(BaseItem):
    """Base class for Things

    :ivar str status: Status of the thing (e.g. OFFLINE, ONLINE, ...)
    """
    def __init__(self, name: str):
        super().__init__(name)

        self.status: str = ''

    def __update_timestamps(self, changed: bool):
        _now = datetime.now(tz=utc)
        self._last_update.set(_now)
        if changed:
            self._last_change.set(_now)
        return None

    def process_event(self, event):

        if isinstance(event, ThingStatusInfoEvent):
            old = self.status
            self.status = event.status
            self.__update_timestamps(old == self.status)

        return None



from HABApp.openhab.items.base_item import OpenhabItem
from ..definitions import QuantityValue


class StringItem(OpenhabItem):
    """StringItem which accepts and converts the data types from OpenHAB"""


class LocationItem(OpenhabItem):
    """LocationItem which accepts and converts the data types from OpenHAB"""

    def set_value(self, new_value) -> bool:

        if isinstance(new_value, QuantityValue):
            return super().set_value(new_value.value)

        return super().set_value(new_value)


class PlayerItem(OpenhabItem):
    """PlayerItem which accepts and converts the data types from OpenHAB"""

    def set_value(self, new_value) -> bool:

        if isinstance(new_value, QuantityValue):
            return super().set_value(new_value.value)

        return super().set_value(new_value)



import asyncio
import logging
import traceback
import typing
from typing import Any, Optional

import aiohttp
from aiohttp.client import ClientResponse, _RequestContextManager
from aiohttp.hdrs import METH_GET, METH_POST, METH_PUT, METH_DELETE
from aiohttp_sse_client import client as sse_client

import HABApp
import HABApp.core
import HABApp.openhab.events
from HABApp.core.const.json import dump_json, load_json
from HABApp.openhab.exceptions import OpenhabConnectionNotSetUpError, OpenhabNotReadyYet, \
    OpenhabDisconnectedError
from .http_connection_waiter import WaitBetweenConnects

log = logging.getLogger('HABApp.openhab.connection')
log_events = logging.getLogger('HABApp.EventBus.openhab')


IS_ONLINE = False
IS_READ_ONLY = False

IS_OH2 = False

HTTP_PREFIX: Optional[str] = None

# HTTP options
HTTP_ALLOW_REDIRECTS: bool = True
HTTP_SESSION: aiohttp.ClientSession = None

CONNECT_WAIT: WaitBetweenConnects = WaitBetweenConnects()


FUT_UUID: Optional[asyncio.Future] = None
FUT_SSE: Optional[asyncio.Future] = None


ON_CONNECTED: typing.Callable = None
ON_DISCONNECTED: typing.Callable = None
ON_SSE_EVENT: typing.Callable[[typing.Dict[str, Any]], Any] = None


async def get(url: str, log_404=True, **kwargs: Any) -> ClientResponse:
    if HTTP_PREFIX is None:
        raise OpenhabConnectionNotSetUpError()

    assert not url.startswith('/'), url
    url = f'{HTTP_PREFIX}/rest/{url}/'

    mgr = _RequestContextManager(HTTP_SESSION._request(METH_GET, url, allow_redirects=HTTP_ALLOW_REDIRECTS, **kwargs))
    return await check_response(mgr, log_404=log_404)


async def post(url: str, log_404=True, json=None, data=None, **kwargs: Any) -> Optional[ClientResponse]:
    if HTTP_PREFIX is None:
        raise OpenhabConnectionNotSetUpError()

    if IS_READ_ONLY or not IS_ONLINE:
        return None

    assert not url.startswith('/'), url
    url = f'{HTTP_PREFIX}/rest/{url}/'

    mgr = _RequestContextManager(
        HTTP_SESSION._request(METH_POST, url, allow_redirects=HTTP_ALLOW_REDIRECTS, data=data, json=json, **kwargs)
    )

    if data is None:
        data = json
    return await check_response(mgr, log_404=log_404, sent_data=data)


async def put(url: str, log_404=True, json=None, data=None, **kwargs: Any) -> Optional[ClientResponse]:
    if HTTP_PREFIX is None:
        raise OpenhabConnectionNotSetUpError()

    if IS_READ_ONLY or not IS_ONLINE:
        return None

    assert not url.startswith('/'), url
    url = f'{HTTP_PREFIX}/rest/{url}/'

    mgr = _RequestContextManager(
        HTTP_SESSION._request(METH_PUT, url, allow_redirects=HTTP_ALLOW_REDIRECTS, data=data, json=json, **kwargs)
    )

    if data is None:
        data = json
    return await check_response(mgr, log_404=log_404, sent_data=data)


async def delete(url: str, log_404=True, json=None, data=None, **kwargs: Any) -> Optional[ClientResponse]:
    if HTTP_PREFIX is None:
        raise OpenhabConnectionNotSetUpError()

    if IS_READ_ONLY or not IS_ONLINE:
        return None

    assert not url.startswith('/'), url
    url = f'{HTTP_PREFIX}/rest/{url}/'

    mgr = _RequestContextManager(
        HTTP_SESSION._request(METH_DELETE, url, allow_redirects=HTTP_ALLOW_REDIRECTS, data=data, json=json, **kwargs)
    )

    if data is None:
        data = json
    return await check_response(mgr, log_404=log_404, sent_data=data)


def set_offline(log_msg=''):
    global IS_ONLINE, FUT_UUID, FUT_SSE

    if not IS_ONLINE:
        return None
    IS_ONLINE = False

    log.warning(f'Disconnected! {log_msg}')

    # cancel SSE listener
    if FUT_SSE is not None:
        if not FUT_SSE.done():
            FUT_SSE.cancel()
        FUT_SSE = None

    ON_DISCONNECTED()

    # Try reconnect
    if not FUT_UUID.done():
        FUT_UUID.cancel()
    FUT_UUID = asyncio.run_coroutine_threadsafe(try_uuid(), HABApp.core.const.loop)


def is_disconnect_exception(e) -> bool:
    if not isinstance(e, (
            # aiohttp Exceptions
            aiohttp.ClientPayloadError, aiohttp.ClientConnectorError, aiohttp.ClientOSError,

            # aiohttp_sse_client Exceptions
            ConnectionRefusedError, ConnectionError, ConnectionAbortedError)):
        return False

    set_offline(str(e))
    return True


async def check_response(future: aiohttp.client._RequestContextManager, sent_data=None,
                         log_404=True) -> ClientResponse:
    try:
        resp = await future
    except Exception as e:
        is_disconnect = is_disconnect_exception(e)
        log.log(logging.WARNING if is_disconnect else logging.ERROR, f'"{e}" ({type(e)})')
        if is_disconnect:
            raise OpenhabDisconnectedError()
        raise

    status = resp.status

    # Server Errors if openhab is not ready yet
    if status >= 500:
        set_offline(f'Status {status} for {resp.request_info.method} {resp.request_info.url}')
        raise OpenhabNotReadyYet()

    # Something went wrong - log error message
    log_msg = False
    if status >= 300:
        log_msg = True

        # possibility skip logging of 404
        if status == 404 and not log_404:
            log_msg = False

    if log_msg:
        # Log Error Message
        sent = '' if sent_data is None else f' {sent_data}'
        log.warning(f'Status {status} for {resp.request_info.method} {resp.request_info.url}{sent}')
        for line in str(resp).splitlines():
            log.debug(line)

    return resp


def stop_connection():
    global FUT_UUID, FUT_SSE
    if FUT_UUID is not None and not FUT_UUID.done():
        FUT_UUID.cancel()
        FUT_UUID = None

    if FUT_SSE is not None and not FUT_SSE.done():
        FUT_SSE.cancel()
        FUT_SSE = None


async def start_connection():
    global HTTP_PREFIX, HTTP_SESSION, FUT_UUID

    stop_connection()

    # If we are already connected properly disconnect
    if HTTP_SESSION is not None:
        await HTTP_SESSION.close()
        HTTP_SESSION = None

    host: str = HABApp.CONFIG.openhab.connection.host
    port: str = HABApp.CONFIG.openhab.connection.port

    # do not run without host
    if host == '':
        HTTP_PREFIX = None
        return None

    HTTP_PREFIX = f'http://{host:s}:{port:d}'

    auth = None
    if HABApp.CONFIG.openhab.connection.user or HABApp.CONFIG.openhab.connection.password:
        auth = aiohttp.BasicAuth(
            HABApp.CONFIG.openhab.connection.user,
            HABApp.CONFIG.openhab.connection.password
        )

    # todo: add possibility to configure line size with read_bufsize
    HTTP_SESSION = aiohttp.ClientSession(
        timeout=aiohttp.ClientTimeout(total=None),
        json_serialize=dump_json,
        auth=auth,
        read_bufsize=2**19  # 512k buffer
    )

    FUT_UUID = asyncio.ensure_future(try_uuid())


async def start_sse_event_listener():
    try:
        # cache so we don't have to look up every event
        call = ON_SSE_EVENT

        event_prefix = 'openhab' if not IS_OH2 else 'smarthome'

        async with sse_client.EventSource(
                url=f'{HTTP_PREFIX}/rest/events?topics='
                    f'{event_prefix}/items/,'                   # Item updates
                    f'{event_prefix}/channels/,'                # Channel update
                    f'{event_prefix}/things/*/status,'          # Thing status updates
                    f'{event_prefix}/things/*/statuschanged'    # Thing status changes
                ,
                session=HTTP_SESSION
        ) as event_source:
            async for event in event_source:
                try:
                    event = load_json(event.data)
                except ValueError:
                    continue
                except TypeError:
                    continue

                # Log sse event
                if log_events.isEnabledFor(logging.DEBUG):
                    log_events._log(logging.DEBUG, event, [])

                # process
                call(event)

    except asyncio.CancelledError:
        # This exception gets raised if we cancel the coroutine
        # since this is normal behaviour we ignore this exception
        pass
    except Exception as e:
        disconnect = is_disconnect_exception(e)
        lvl = logging.WARNING if disconnect else logging.ERROR
        log.log(lvl, f'SSE request Error: {e}')
        for line in traceback.format_exc().splitlines():
            log.log(lvl, line)

        # reconnect even if we have an unexpected error
        if not disconnect:
            set_offline(f'Uncaught error in process_sse_events: {e}')


async def async_get_uuid() -> str:
    resp = await get('uuid', log_404=False)
    if resp.status >= 300:
        raise OpenhabNotReadyYet()
    return await resp.text(encoding='utf-8')


async def async_get_root() -> dict:
    resp = await get('', log_404=False)
    if resp.status == 404:
        return {}
    return await resp.json(encoding='utf-8')


def patch_for_oh2(reverse=False):
    global IS_OH2

    IS_OH2 = True

    # events are named different
    HABApp.openhab.events.item_events.NAME_START = 16 if not reverse else 14
    HABApp.openhab.events.thing_events.NAME_START = 17 if not reverse else 15
    HABApp.openhab.events.channel_events.NAME_START = 19 if not reverse else 17


async def try_uuid():
    global FUT_UUID, FUT_SSE, IS_ONLINE

    # sleep before reconnect
    await CONNECT_WAIT.wait()

    log.debug('Trying to connect to OpenHAB ...')
    try:
        uuid = await async_get_uuid()
        root = await async_get_root()      # this will only work on OH3
    except Exception as e:
        if isinstance(e, (OpenhabDisconnectedError, OpenhabNotReadyYet)):
            log.info('... offline!')
        else:
            for line in traceback.format_exc().splitlines():
                log.error(line)

        # Keep trying to connect
        FUT_UUID = asyncio.ensure_future(try_uuid())
        return None

    if IS_READ_ONLY:
        log.info(f'Connected read only to OpenHAB instance {uuid}')
    else:
        log.info(f'Connected to OpenHAB instance {uuid}')

    info = root.get('runtimeInfo')
    if info is None:
        patch_for_oh2()
    else:
        log.info(f'OpenHAB version {info["version"]} ({info["buildString"]})')

    IS_ONLINE = True

    # start sse processing
    if FUT_SSE is not None:
        FUT_SSE.cancel()
    FUT_SSE = asyncio.ensure_future(start_sse_event_listener())

    ON_CONNECTED()
    return None


def __load_cfg():
    global IS_READ_ONLY
    IS_READ_ONLY = HABApp.config.CONFIG.openhab.general.listen_only


# setup config
__load_cfg()
HABApp.config.CONFIG.subscribe_for_changes(__load_cfg)



import logging
import typing

import paho.mqtt.client as mqtt

import HABApp
from HABApp.core.wrapper import log_exception

from HABApp.runtime.shutdown_helper import ShutdownHelper
from ..config import Mqtt as MqttConfig
from ..core.const.json import load_json

from .events import MqttValueUpdateEvent, MqttValueChangeEvent

log = logging.getLogger('HABApp.mqtt.connection')
log_msg = logging.getLogger('HABApp.EventBus.mqtt')


class MqttConnection:
    def __init__(self, mqtt_config: MqttConfig, shutdown_helper: ShutdownHelper):
        assert isinstance(mqtt_config, MqttConfig)

        self.loop_started = False
        self.connected = False

        self.client: mqtt.Client = None

        self.subscriptions: typing.List[typing.Tuple[str, int]] = []

        # config changes
        self.__config = mqtt_config
        self.__config.subscribe.subscribe_for_changes(self.subscription_changed)
        self.__config.connection.subscribe_for_changes(self.connect)

        # shutdown
        shutdown_helper.register_func(self.disconnect)

        self.interface: HABApp.mqtt.MqttInterface = HABApp.mqtt.mqtt_interface.get_mqtt_interface(self, self.__config)

    def connect(self):
        if not self.__config.connection.host:
            log.info('MQTT disabled')
            self.disconnect()
            return None

        if self.connected:
            log.info('disconnecting')
            self.client.disconnect()
            self.connected = False

        self.client = mqtt.Client(
            client_id=self.__config.connection.client_id,
            clean_session=False
        )

        if self.__config.connection.tls:
            self.client.tls_set()

            # we can only set tls_insecure if we have a tls connection
            if self.__config.connection.tls_insecure:
                log.warning('Verification of server hostname in server certificate disabled!')
                log.warning('Use this only for testing, not for a real system!')
                self.client.tls_insecure_set(True)

        # set user if required
        user = self.__config.connection.user
        pw = self.__config.connection.password
        if user:
            self.client.username_pw_set(
                user,
                pw if pw else None
            )

        # setup callbacks
        self.client.on_connect = self.on_connect
        self.client.on_disconnect = self.on_disconnect
        self.client.on_message = self.process_msg

        self.client.connect_async(
            self.__config.connection.host,
            port=self.__config.connection.port,
            keepalive=60
        )

        log.info(f'Connecting to {self.__config.connection.host}:{self.__config.connection.port}')

        if not self.loop_started:
            self.client.loop_start()
        self.loop_started = True

    @log_exception
    def subscription_changed(self):
        if not self.connected:
            return None

        if self.subscriptions:
            unsubscribe = [k[0] for k in self.subscriptions]
            log.debug('Unsubscribing from:')
            for t in unsubscribe:
                log.debug(f' - "{t}"')
            self.client.unsubscribe(unsubscribe)

        topics = self.__config.subscribe.topics
        default_qos = self.__config.subscribe.qos
        self.subscriptions = [(topic, qos if qos is not None else default_qos) for topic, qos in topics]
        log.debug('Subscribing to:')
        for topic, qos in self.subscriptions:
            log.debug(f' - "{topic}" (QoS {qos:d})')
        self.client.subscribe(self.subscriptions)

    @log_exception
    def on_connect(self, client, userdata, flags, rc):
        log.log(logging.INFO if not rc else logging.ERROR, mqtt.connack_string(rc))
        if rc:
            return None
        self.connected = True

        self.subscriptions.clear()
        self.subscription_changed()

    @log_exception
    def on_disconnect(self, client, userdata, rc):
        log.log(logging.INFO if not rc else logging.ERROR, f'Disconnect: {mqtt.error_string(rc)} ({rc})')
        self.connected = False

    @log_exception
    def process_msg(self, client, userdata, message: mqtt.MQTTMessage):
        topic: str = message.topic

        try:
            payload = message.payload.decode("utf-8")

            if log_msg.isEnabledFor(logging.DEBUG):
                log_msg._log(logging.DEBUG, f'{topic} ({message.qos}): {payload}', [])

            # load json dict and list
            if payload.startswith('{') and payload.endswith('}') or payload.startswith('[') and payload.endswith(']'):
                try:
                    payload = load_json(payload)
                except ValueError:
                    pass
            else:
                # try to cast to int/float
                try:
                    payload = int(payload)
                except ValueError:
                    try:
                        payload = float(payload)
                    except ValueError:
                        pass
        except UnicodeDecodeError:
            # Payload ist binary
            payload = message.payload
            if log_msg.isEnabledFor(logging.DEBUG):
                log_msg._log(logging.DEBUG, f'{topic} ({message.qos}): {payload[:20]}...', [])


        # get the mqtt item
        _item = HABApp.mqtt.items.MqttItem.get_create_item(topic)

        # remeber state and update item before doing callbacks
        _old_state = _item.value
        _item.set_value(payload)

        # Post events
        HABApp.core.EventBus.post_event(topic, MqttValueUpdateEvent(topic, payload))
        if _old_state != payload:
            HABApp.core.EventBus.post_event(topic, MqttValueChangeEvent(topic, payload, _old_state))

    def disconnect(self):
        if self.connected:
            self.client.disconnect()
            self.connected = False
        if self.loop_started:
            self.client.loop_stop()
            self.loop_started = False

        self.client = None



import typing

from HABApp.core.items.base_item import BaseItem as __BaseItem

_ALL_ITEMS: typing.Dict[str, __BaseItem] = {}


class ItemNotFoundException(Exception):
    def __init__(self, name: str):
        super().__init__(f'Item {name} does not exist!')
        self.name: str = name


class ItemAlreadyExistsError(Exception):
    def __init__(self, name: str):
        super().__init__(f'Item {name} does already exist and can not be added again!')
        self.name: str = name


def item_exists(name: str) -> bool:
    return name in _ALL_ITEMS


def get_item(name: str) -> __BaseItem:
    try:
        return _ALL_ITEMS[name]
    except KeyError:
        raise ItemNotFoundException(name) from None


def get_all_items() -> typing.List[__BaseItem]:
    return list(_ALL_ITEMS.values())


def get_all_item_names() -> typing.List[str]:
    return list(_ALL_ITEMS.keys())


def create_item(name: str, item_factory, initial_value=None) -> __BaseItem:
    assert issubclass(item_factory, __BaseItem), item_factory
    new_item = item_factory(name, initial_value=initial_value)
    add_item(new_item)
    return new_item


def add_item(item: __BaseItem):
    assert isinstance(item, __BaseItem), type(item)
    name = item.name

    existing = _ALL_ITEMS.get(name)
    if existing is not None:
        # adding the same item multiple times will not cause an exception
        if existing is item:
            return None

        # adding a new item with the same name raises an exception
        raise ItemAlreadyExistsError(name)

    _ALL_ITEMS[name] = item
    item._on_item_add()


def pop_item(name: str) -> __BaseItem:
    try:
        item = _ALL_ITEMS.pop(name)
    except KeyError:
        raise ItemNotFoundException(name) from None

    item._on_item_remove()
    return item



import asyncio
from pathlib import Path
from typing import Dict, Set, List

import HABApp
from HABApp.core.lib import PendingFuture
from HABApp.core.logger import log_warning, HABAppError
from HABApp.openhab.connection_handler.func_async import async_get_things
from HABApp.openhab.connection_logic.plugin_things.cfg_validator import validate_cfg, InvalidItemNameError
from HABApp.openhab.connection_logic.plugin_things.filters import THING_ALIAS, CHANNEL_ALIAS
from HABApp.openhab.connection_logic.plugin_things.filters import apply_filters, log_overview
from ._log import log
from .item_worker import create_item, cleanup_items
from .items_file import create_items_file
from .thing_worker import update_thing_cfg
from .._plugin import OnConnectPlugin


class ManualThingConfig(OnConnectPlugin):

    def __init__(self):
        super().__init__()
        self.created_items: Dict[str, Set[str]] = {}
        self.do_cleanup = PendingFuture(self.clean_items, 120)

    def setup(self):
        if not HABApp.CONFIG.directories.config.is_dir():
            log.info('Config folder does not exist - textual thing config disabled!')
            return None

        # Add event bus listener
        HABApp.core.EventBus.add_listener(
            HABApp.core.EventBusListener(
                HABApp.core.const.topics.FILES,
                HABApp.core.WrappedFunction(self.file_load_event),
                HABApp.core.events.habapp_events.RequestFileLoadEvent
            )
        )

        # watch folder
        HABApp.core.files.watch_folder(HABApp.CONFIG.directories.config, '.yml', True)

    async def file_load_event(self, event: HABApp.core.events.habapp_events.RequestFileLoadEvent):
        if HABApp.core.files.file_name.is_config(event.name):
            await self.update_thing_config(event.get_path())

    async def on_connect_function(self):
        try:
            await asyncio.sleep(0.3)

            files = list(HABApp.core.lib.list_files(HABApp.CONFIG.directories.config, '.yml'))
            if not files:
                log.debug(f'No manual configuration files found in {HABApp.CONFIG.directories.config}')
                return None

            # if oh is not ready we will get None, but we will trigger again on reconnect
            data = await async_get_things()
            if data is None:
                return None

            for f in files:
                await self.update_thing_config(f, data)
        except asyncio.CancelledError:
            pass

    @HABApp.core.wrapper.ignore_exception
    async def clean_items(self):
        items = set()
        for s in self.created_items.values():
            items.update(s)
        await cleanup_items(items)

    async def update_thing_configs(self, files: List[Path]):
        data = await async_get_things()
        if data is None:
            return None

        for file in files:
            await self.update_thing_config(file, data)

    @HABApp.core.wrapper.ignore_exception
    async def update_thing_config(self, path: Path, data=None):
        # we have to check the naming structure because we get file events for the whole folder
        _name = path.name.lower()
        if not _name.startswith('thing_') or not _name.endswith('.yml'):
            return None

        # only load if we don't supply the data
        if data is None:
            data = await async_get_things()

        # remove created items
        self.created_items.pop(path.name, None)
        created_items = self.created_items.setdefault(path.name, set())

        # shedule cleanup
        self.do_cleanup.reset()

        # output file
        output_file = path.with_suffix('.items')
        if output_file.is_file():
            output_file.unlink()

        # we also get events when the file gets deleted
        if not path.is_file():
            log.debug(f'File {path} does not exist -> skipping Thing configuration!')
            return None
        log.debug(f'Loading {path}!')

        # load the config file
        with path.open(mode='r', encoding='utf-8') as file:
            try:
                cfg = HABApp.core.const.yml.load(file)
            except Exception as e:
                HABAppError(log).add_exception(e).dump()
                return None

        # validate configuration
        cfg = validate_cfg(cfg, path.name)
        if cfg is None:
            return None

        # if one entry has test set we show an overview of all the things
        if any(map(lambda x: x.test, cfg)):
            log_overview(data, THING_ALIAS, 'Thing overview')

        # process each thing part in the cfg
        for cfg_entry in cfg:
            test: bool = cfg_entry.test
            things = list(apply_filters(cfg_entry.filter, data, test))

            # show a warning we found no Things
            if not things:
                log_warning(log, f'No things matched for {cfg_entry.filter}')
                continue

            # update thing configuration
            if cfg_entry.thing_config:
                await update_thing_cfg(cfg_entry.thing_config, things, test)

            try:
                # item creation for every thing
                create_items = {}
                shown_types = set()
                for thing in things:
                    thing_context = {k: thing.get(alias, '') for k, alias in THING_ALIAS.items()}

                    # create items without channel
                    for item_cfg in cfg_entry.get_items(thing_context):
                        name = item_cfg.name
                        if name in create_items:
                            raise ValueError(f'Duplicate item: {name}')
                        create_items[name] = item_cfg

                    # Channel overview, only if we have something configured
                    if test and cfg_entry.channels:
                        __thing_type = thing_context['thing_type']
                        if __thing_type not in shown_types:
                            shown_types.add(__thing_type)
                            log_overview(thing['channels'], CHANNEL_ALIAS, heading=f'Channels for {__thing_type}')

                    # do channel things
                    for channel_cfg in cfg_entry.channels:
                        channels = apply_filters(channel_cfg.filter, thing['channels'], test)
                        for channel in channels:
                            channel_context = {k: channel.get(alias, '') for k, alias in CHANNEL_ALIAS.items()}
                            channel_context.update(thing_context)

                            for item_cfg in channel_cfg.get_items(channel_context):
                                item_cfg.link = channel['uid']
                                name = item_cfg.name

                                if name in create_items:
                                    raise ValueError(f'Duplicate item: {name}')
                                create_items[name] = item_cfg

                    # newline only if we create logs
                    if test and (cfg_entry.create_items or cfg_entry.channels):
                        log.info('')
            except InvalidItemNameError as e:
                HABAppError(log).add_exception(e).dump()
                continue

            # Create all items
            for item_cfg in create_items.values():
                created = await create_item(item_cfg, test)
                if created:
                    created_items.add(item_cfg.name)

            self.do_cleanup.reset()

            create_items_file(output_file, create_items)


PLUGIN_MANUAL_THING_CFG = ManualThingConfig.create_plugin()



from .base import ScheduledCallbackBase, TYPING_DATE_TIME


class OneTimeCallback(ScheduledCallbackBase):

    def _calculate_next_call(self):
        if self.run_counter:
            self.is_finished = True

    def set_run_time(self, next_time: TYPING_DATE_TIME) -> 'OneTimeCallback':
        self._set_next_base(next_time)
        self._update_run_time()
        return self



from .counter_item import CounterItem
from .period_counter import PeriodCounter
from .threshold import Threshold
from .statistics import Statistics
from . import multimode

# 27.04.2020 - this can be removed in some time
from .multimode import MultiModeItem



import asyncio
import datetime
import logging
import math
import threading
import time
import typing

from pytz import utc

import HABApp
from pathlib import Path
from HABApp.core.files import file_load_failed, file_load_ok
from HABApp.core.files.watcher import AggregatingAsyncEventHandler
from HABApp.core.logger import log_warning
from HABApp.core.wrapper import log_exception
from .rule_file import RuleFile

log = logging.getLogger('HABApp.Rules')


LOAD_DELAY = 1


async def set_load_ok(name: str):
    await asyncio.sleep(LOAD_DELAY)
    file_load_ok(name)


async def set_load_failed(name: str):
    await asyncio.sleep(LOAD_DELAY)
    file_load_failed(name)


class RuleManager:

    def __init__(self, parent):
        assert isinstance(parent, HABApp.runtime.Runtime)
        self.runtime = parent

        self.files: typing.Dict[str, RuleFile] = {}

        # serialize loading
        self.__load_lock = threading.Lock()
        self.__files_lock = threading.Lock()

        # Processing
        self.__process_last_sec = 60

        self.watcher: typing.Optional[AggregatingAsyncEventHandler] = None

    def setup(self):

        # Add event bus listener
        HABApp.core.files.add_event_bus_listener('rule', self.request_file_load, self.request_file_unload, log)

        # Folder watcher
        self.watcher = HABApp.core.files.watch_folder(HABApp.CONFIG.directories.rules, '.py', True)

        # Initial loading of rules
        HABApp.core.WrappedFunction(self.load_rules_on_startup, logger=log, warn_too_long=False).run()

    def load_rules_on_startup(self):

        if HABApp.CONFIG.openhab.connection.host and HABApp.CONFIG.openhab.general.wait_for_openhab:
            items_found = False
            while not items_found:
                time.sleep(3)
                for item in HABApp.core.Items.get_all_items():
                    if isinstance(item, HABApp.openhab.items.OpenhabItem):
                        items_found = True
                        break

                # stop waiting if we want to shut down
                if self.runtime.shutdown.requested:
                    return None
            time.sleep(2.2)
        else:
            time.sleep(5.2)

        # trigger event for every file
        self.watcher.trigger_all()
        return None

    @log_exception
    async def process_scheduled_events(self):

        while not self.runtime.shutdown.requested:

            now = datetime.datetime.now(tz=utc)

            # process only once per second
            if now.second == self.__process_last_sec:
                await asyncio.sleep(0.1)
                continue

            # remember sec
            self.__process_last_sec = now.second

            with self.__files_lock:
                for file in self.files.values():
                    assert isinstance(file, RuleFile), type(file)
                    for rule in file.rules.values():
                        rule._process_events(now)

            # sleep longer, try to sleep until the next full second
            end = datetime.datetime.now(tz=utc)
            if end.second == self.__process_last_sec:
                frac, whole = math.modf(time.time())
                sleep_time = 1 - frac + 0.005   # prevent rounding error and add a little bit of security
                await asyncio.sleep(sleep_time)


    @log_exception
    def get_async(self):
        return asyncio.gather(self.process_scheduled_events())

    @log_exception
    def get_rule(self, rule_name):
        found = []
        for file in self.files.values():
            if rule_name is None:
                for rule in file.rules.values():
                    found.append(rule)
            else:
                if rule_name in file.rules:
                    found.append(file.rules[rule_name])

        # if we want all return them
        if rule_name is None:
            return found

        # if we want a special one throw error
        if not found:
            raise KeyError(f'No Rule with name "{rule_name}" found!')
        return found if len(found) > 1 else found[0]


    @log_exception
    def request_file_unload(self, name: str, path: Path, request_lock=True):
        path_str = str(path)

        try:
            if request_lock:
                self.__load_lock.acquire()

            # Only unload already loaded files
            with self.__files_lock:
                already_loaded = path_str in self.files
            if not already_loaded:
                log_warning(log, f'Rule file {path} is not yet loaded and therefore can not be unloaded')
                return None

            log.debug(f'Removing file: {path}')
            with self.__files_lock:
                rule = self.files.pop(path_str)
            rule.unload()
        except Exception as e:
            err = HABApp.core.logger.HABAppError(log)
            err.add(f"Could not remove {path}!")
            err.add_exception(e, True)
            err.dump()
            return None
        finally:
            if request_lock:
                self.__load_lock.release()

    @log_exception
    def request_file_load(self, name: str, path: Path):
        path_str = str(path)

        # Only load existing files
        if not path.is_file():
            log_warning(log, f'Rule file {path} does not exist and can not be loaded!')
            return None

        with self.__load_lock:
            # Unload if we have already loaded
            with self.__files_lock:
                already_loaded = path_str in self.files
            if already_loaded:
                self.request_file_unload(name, path, request_lock=False)

            log.debug(f'Loading file: {path}')
            with self.__files_lock:
                self.files[path_str] = file = RuleFile(self, path)

            if not file.load():
                # If the load has failed we remove it again.
                # Unloading is handled directly in the load function
                self.files.pop(path_str)
                log.warning(f'Failed to load {path_str}!')

                # signal that we have loaded the file but with a small delay
                asyncio.run_coroutine_threadsafe(set_load_failed(name), HABApp.core.const.loop)
                return None

        log.debug(f'File {path_str} successfully loaded!')

        # signal that we have loaded the file but with a small delay
        asyncio.run_coroutine_threadsafe(set_load_ok(name), HABApp.core.const.loop)

        # Do simple checks which prevent errors
        file.check_all_rules()



import typing
from pathlib import Path

import setuptools  # type: ignore


# Load version number without importing HABApp
def load_version() -> str:
    version: typing.Dict[str, str] = {}
    with open("HABApp/__version__.py") as fp:
        exec(fp.read(), version)
    assert version['__version__'], version
    return version['__version__']


def load_req() -> typing.List[str]:
    # When we run tox tests we don't have this file available so we skip them
    req_file = Path(__file__).with_name('requirements.txt')
    if not req_file.is_file():
        return ['']

    with req_file.open() as f:
        return f.readlines()


__version__ = load_version()

print(f'Version: {__version__}')
print('')

# When we run tox tests we don't have these files available so we skip them
readme = Path(__file__).with_name('readme.md')
long_description = ''
if readme.is_file():
    with readme.open("r", encoding='utf-8') as fh:
        long_description = fh.read()

setuptools.setup(
    name="HABApp",
    version=__version__,
    author="spaceman_spiff",
    # author_email="",
    description="Easy automation with MQTT and/or openHAB. Create home automation rules in python.",
    keywords=[
        'mqtt',
        'openhab',
        'habapp',
        'home automation'
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/spacemanspiff2007/HABApp",
    project_urls={
        'Documentation': 'https://habapp.readthedocs.io/',
        'GitHub': 'https://github.com/spacemanspiff2007/HABApp',
    },
    packages=setuptools.find_packages(exclude=['tests*']),
    install_requires=load_req(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: AsyncIO",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Home Automation"
    ],
    entry_points={
        'console_scripts': [
            'habapp = HABApp.__main__:main'
        ]
    }
)



import HABApp
import HABApp.core
import HABApp.openhab.events
from HABApp.core import Items
from HABApp.core.wrapper import ignore_exception
from HABApp.openhab.connection_handler import http_connection
from HABApp.openhab.map_events import get_event
from HABApp.openhab.map_items import map_item
from ._plugin import on_connect, on_disconnect, setup_plugins

log = http_connection.log


def setup(shutdown):
    assert isinstance(shutdown, HABApp.runtime.ShutdownHelper), type(shutdown)

    # initialize callbacks
    http_connection.ON_CONNECTED = on_connect
    http_connection.ON_DISCONNECTED = on_disconnect
    http_connection.ON_SSE_EVENT = on_sse_event

    # shutdown handler for connection
    shutdown.register_func(http_connection.stop_connection)

    # shutdown handler for plugins
    shutdown.register_func(on_disconnect)

    # initialize all plugins
    setup_plugins()
    return None


async def start():
    await http_connection.start_connection(),


@ignore_exception
def on_sse_event(event_dict: dict):

    # Lookup corresponding OpenHAB event
    event = get_event(event_dict)

    # Update item in registry BEFORE posting to the event bus
    # so the items have the correct state when we process the event in a rule
    try:
        if isinstance(event, HABApp.core.events.ValueUpdateEvent):
            __item = Items.get_item(event.name)  # type: HABApp.core.items.base_item.BaseValueItem
            __item.set_value(event.value)
            HABApp.core.EventBus.post_event(event.name, event)
            return None
        elif isinstance(event, HABApp.openhab.events.ThingStatusInfoEvent):
            __thing = Items.get_item(event.name)   # type: HABApp.openhab.items.Thing
            __thing.process_event(event)
            HABApp.core.EventBus.post_event(event.name, event)
            return None
    except HABApp.core.Items.ItemNotFoundException:
        pass

    # Events which change the ItemRegistry
    if isinstance(event, (HABApp.openhab.events.ItemAddedEvent, HABApp.openhab.events.ItemUpdatedEvent)):
        item = map_item(event.name, event.type, 'NULL')
        if item is None:
            return None

        # check already existing item so we can print a warning if something changes
        try:
            existing_item = Items.get_item(item.name)
            if isinstance(existing_item, item.__class__):
                # it's the same item class so we don't replace it!
                item = existing_item
            else:
                log.warning(f'Item changed type from {existing_item.__class__} to {item.__class__}')
                # remove the item so it can be added again
                Items.pop_item(item.name)
        except Items.ItemNotFoundException:
            pass

        # always overwrite with new definition
        Items.add_item(item)

    elif isinstance(event, HABApp.openhab.events.ItemRemovedEvent):
        Items.pop_item(event.name)

    # Send Event to Event Bus
    HABApp.core.EventBus.post_event(event.name, event)
    return None



import asyncio
import datetime
import logging
import random
import sys
import traceback
import typing
import warnings
import weakref

import HABApp
import HABApp.core
import HABApp.openhab
import HABApp.rule_manager
import HABApp.util
from HABApp.core.events import AllEvents
from .interfaces import async_subprocess_exec
from .scheduler import ReoccurringScheduledCallback, OneTimeCallback, DayOfWeekScheduledCallback, \
    TYPING_DATE_TIME, SunScheduledCallback
from .scheduler.base import ScheduledCallbackBase as _ScheduledCallbackBase


log = logging.getLogger('HABApp.Rule')


# Func to log deprecation warnings
def send_warnings_to_log(message, category, filename, lineno, file=None, line=None):
    log.warning('%s:%s: %s:%s' % (filename, lineno, category.__name__, message))
    return


# Setup deprecation warnings
warnings.simplefilter('default')
warnings.showwarning = send_warnings_to_log


class Rule:
    def __init__(self):

        # get the variables from the caller
        depth = 1
        while True:
            __vars = sys._getframe(depth).f_globals
            depth += 1
            if '__HABAPP__RUNTIME__' in __vars:
                __runtime__ = __vars['__HABAPP__RUNTIME__']
                __rule_file__ = __vars['__HABAPP__RULE_FILE__']
                break

        # variable vor unittests
        test = __vars.get('__UNITTEST__', False)

        # this is a list which contains all rules of this file
        __vars['__HABAPP__RULES'].append(self)

        assert isinstance(__runtime__, HABApp.runtime.Runtime)
        self.__runtime: HABApp.runtime.Runtime = __runtime__

        if not test:
            assert isinstance(__rule_file__, HABApp.rule_manager.RuleFile)
        self.__rule_file: HABApp.rule_manager.RuleFile = __rule_file__

        self.__event_listener: typing.List[HABApp.core.EventBusListener] = []
        self.__future_events: typing.List[_ScheduledCallbackBase] = []
        self.__unload_functions: typing.List[typing.Callable[[], None]] = []
        self.__cancel_objs: weakref.WeakSet = weakref.WeakSet()

        # schedule cleanup of this rule
        self.register_on_unload(self.__cleanup_rule)
        self.register_on_unload(self.__cleanup_objs)

        # suggest a rule name if it is not
        self.rule_name: str = self.__rule_file.suggest_rule_name(self)

        # interfaces
        self.async_http: HABApp.rule.interfaces.AsyncHttpConnection = self.__runtime.async_http if not test else None
        self.mqtt = self.__runtime.mqtt_connection.interface if not test else None
        self.oh = HABApp.openhab.interface
        self.openhab = self.oh

    @HABApp.core.wrapper.log_exception
    def __cleanup_objs(self):
        while self.__cancel_objs:
            # we log each error as warning
            with HABApp.core.wrapper.ExceptionToHABApp(log, logging.WARNING):
                obj = self.__cancel_objs.pop()
                obj.cancel()

    @HABApp.core.wrapper.log_exception
    def __cleanup_rule(self):
        # Important: set the dicts to None so we don't schedule a future event during _cleanup.
        # If dict is set to None we will crash instead but it is no problem because everything gets unloaded anyhow
        event_listeners = self.__event_listener
        future_events = self.__future_events

        self.__event_listener = None
        self.__future_events = None

        # Actually remove the listeners/events
        for listener in event_listeners:
            HABApp.core.EventBus.remove_listener(listener)

        for event in future_events:
            event.cancel()
        return None

    def post_event(self, name, event):
        """
        Post an event to the event bus

        :param name: name or item to post event to
        :param event: Event class to be used (must be class instance)
        :return:
        """
        assert isinstance(name, (str, HABApp.core.items.BaseValueItem)), type(name)
        return HABApp.core.EventBus.post_event(
            name.name if isinstance(name, HABApp.core.items.BaseValueItem) else name,
            event
        )

    def listen_event(self, name: typing.Union[HABApp.core.items.BaseValueItem, str],
                     callback: typing.Callable[[typing.Any], typing.Any],
                     event_type: typing.Union[AllEvents, typing.Any] = AllEvents
                     ) -> HABApp.core.EventBusListener:
        """
        Register an event listener

        :param name: item or name to listen to. Use None to listen to all events
        :param callback: callback that accepts one parameter which will contain the event
        :param event_type: Event filter. This is typically :class:`~HABApp.core.ValueUpdateEvent` or
            :class:`~HABApp.core.ValueChangeEvent` which will also trigger on changes/update from openhab
            or mqtt.
        """
        cb = HABApp.core.WrappedFunction(callback, name=self._get_cb_name(callback))
        listener = HABApp.core.EventBusListener(
            name.name if isinstance(name, HABApp.core.items.BaseValueItem) else name, cb, event_type
        )
        self.__event_listener.append(listener)
        HABApp.core.EventBus.add_listener(listener)
        return listener

    def execute_subprocess(self, callback, program, *args, capture_output=True):
        """Run another program

        :param callback: |param_scheduled_cb| after process has finished. First parameter will
                         be an instance of :class:`~HABApp.rule.FinishedProcessInfo`
        :param program: program or path to program to run
        :param args: |param_scheduled_cb_args|
        :param capture_output: Capture program output, set to `False` to only capture return code
        :return:
        """

        assert isinstance(program, str), type(program)
        cb = HABApp.core.WrappedFunction(callback, name=self._get_cb_name(callback))

        asyncio.run_coroutine_threadsafe(
            async_subprocess_exec(cb.run, program, *args, capture_output=capture_output),
            HABApp.core.const.loop
        )

    def run_every(self,
                  time: TYPING_DATE_TIME, interval: typing.Union[int, datetime.timedelta],
                  callback, *args, **kwargs) -> ReoccurringScheduledCallback:
        """
        Run a function periodically

        :param time: |param_scheduled_time|
        :param interval:
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        cb = HABApp.core.WrappedFunction(callback, name=self._get_cb_name(callback))
        future_event = ReoccurringScheduledCallback(cb, *args, **kwargs)
        future_event.interval(interval)
        self.__future_events.append(future_event)
        return future_event

    def run_on_sun(self, sun_event: str, callback, *args, run_if_missed=False, **kwargs) -> SunScheduledCallback:
        """Run a function on sunrise/sunset etc

        :param sun_event: 'sunrise', 'sunset', 'dusk', 'dawn'
        :param run_if_missed: run the event if we missed it for today
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        cb = HABApp.core.WrappedFunction(callback, name=self._get_cb_name(callback))
        future_event = SunScheduledCallback(cb, *args, **kwargs)
        future_event.sun_trigger(sun_event)
        future_event._calculate_next_call()
        self.__future_events.append(future_event)
        return future_event

    def run_on_day_of_week(self,
                           time: datetime.time, weekdays, callback, *args, **kwargs) -> DayOfWeekScheduledCallback:
        """

        :param time: datetime.time
        :param weekdays:
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        assert isinstance(time, datetime.time), type(time)

        # names of weekdays in local language
        lookup = {datetime.date(2001, 1, i).strftime('%A'): i for i in range(1, 8)}
        lookup.update({datetime.date(2001, 1, i).strftime('%A')[:3]: i for i in range(1, 8)})

        # abbreviations in German and English
        lookup.update({"Mo": 1, "Di": 2, "Mi": 3, "Do": 4, "Fr": 5, "Sa": 6, "So": 7})
        lookup.update({"Mon": 1, "Tue": 2, "Wed": 3, "Thu": 4, "Fri": 5, "Sat": 6, "Sun": 7})
        lookup = {k.lower(): v for k, v in lookup.items()}

        if isinstance(weekdays, int) or isinstance(weekdays, str):
            weekdays = [weekdays]
        for i, val in enumerate(weekdays):
            if not isinstance(val, str):
                continue
            weekdays[i] = lookup[val.lower()]

        cb = HABApp.core.WrappedFunction(callback, name=self._get_cb_name(callback))
        future_event = DayOfWeekScheduledCallback(cb, *args, **kwargs)
        future_event.weekdays(weekdays)
        future_event.time(time)
        self.__future_events.append(future_event)
        return future_event

    def run_on_every_day(self, time: datetime.time, callback, *args, **kwargs) -> DayOfWeekScheduledCallback:
        """

        :param time: datetime.time
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        assert isinstance(time, datetime.time), type(time)
        cb = HABApp.core.WrappedFunction(callback, name=self._get_cb_name(callback))
        future_event = DayOfWeekScheduledCallback(cb, *args, **kwargs)
        future_event.weekdays('all')
        future_event.time(time)
        self.__future_events.append(future_event)
        return future_event

    def run_on_workdays(self, time: datetime.time, callback, *args, **kwargs) -> DayOfWeekScheduledCallback:
        """

        :param time: datetime.time
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        assert isinstance(time, datetime.time), type(time)
        cb = HABApp.core.WrappedFunction(callback, name=self._get_cb_name(callback))
        future_event = DayOfWeekScheduledCallback(cb, *args, **kwargs)
        future_event.weekdays('workday')
        future_event.time(time)
        self.__future_events.append(future_event)
        return future_event

    def run_on_weekends(self, time: datetime.time, callback, *args, **kwargs) -> DayOfWeekScheduledCallback:
        """

        :param time: datetime.time
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        assert isinstance(time, datetime.time), type(time)
        cb = HABApp.core.WrappedFunction(callback, name=self._get_cb_name(callback))
        future_event = DayOfWeekScheduledCallback(cb, *args, **kwargs)
        future_event.weekdays('weekend')
        future_event.time(time)
        self.__future_events.append(future_event)
        return future_event

    def run_daily(self, callback, *args, **kwargs) -> ReoccurringScheduledCallback:
        """
        Picks a random hour, minute and second and runs the callback every day

        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        start = datetime.timedelta(seconds=random.randint(0, 24 * 3600 - 1))
        interval = datetime.timedelta(days=1)
        return self.run_every(start, interval, callback, *args, **kwargs)

    def run_hourly(self, callback, *args, **kwargs) -> ReoccurringScheduledCallback:
        """
        Picks a random minute and second and run the callback every hour

        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        start = datetime.timedelta(seconds=random.randint(0, 3600 - 1))
        interval = datetime.timedelta(seconds=3600)
        return self.run_every(start, interval, callback, *args, **kwargs)

    def run_minutely(self, callback, *args, **kwargs) -> ReoccurringScheduledCallback:
        """
        Picks a random second and runs the callback every minute

        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        start = datetime.timedelta(seconds=random.randint(0, 60 - 1))
        interval = datetime.timedelta(seconds=60)
        return self.run_every(start, interval, callback, *args, **kwargs)

    def run_at(self, date_time: TYPING_DATE_TIME, callback, *args, **kwargs) -> OneTimeCallback:
        """
        Run a function at a specified date_time

        :param date_time:
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        cb = HABApp.core.WrappedFunction(callback, name=self._get_cb_name(callback))
        future_event = OneTimeCallback(cb, *args, **kwargs)
        future_event.set_run_time(date_time)
        self.__future_events.append(future_event)
        return future_event

    def run_in(self, seconds: typing.Union[int, datetime.timedelta], callback, *args, **kwargs) -> OneTimeCallback:
        """
        Run the callback in x seconds

        :param int seconds: Wait time in seconds or a timedelta obj before calling the function
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        assert isinstance(seconds, (int, datetime.timedelta)), f'{seconds} ({type(seconds)})'
        fut = datetime.timedelta(seconds=seconds) if not isinstance(seconds, datetime.timedelta) else seconds

        cb = HABApp.core.WrappedFunction(callback, name=self._get_cb_name(callback))
        future_event = OneTimeCallback(cb, *args, **kwargs)
        future_event.set_run_time(fut)
        self.__future_events.append(future_event)
        return future_event

    def run_soon(self, callback, *args, **kwargs) -> OneTimeCallback:
        """
        Run the callback as soon as possible (typically in the next second).

        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        cb = HABApp.core.WrappedFunction(callback, name=self._get_cb_name(callback))
        future_event = OneTimeCallback(cb, *args, **kwargs)
        future_event.set_run_time(None)
        self.__future_events.append(future_event)
        return future_event

    def get_rule(self, rule_name: str) -> 'typing.Union[Rule, typing.List[Rule]]':
        assert rule_name is None or isinstance(rule_name, str), type(rule_name)
        return self.__runtime.rule_manager.get_rule(rule_name)

    def register_on_unload(self, func: typing.Callable[[], typing.Any]):
        """Register a function with no parameters which will be called when the rule is unloaded.
        Use this for custom cleanup functions.

        :param func: function which will be called
        """
        assert callable(func)
        assert func not in self.__unload_functions, 'Function was already registered!'
        self.__unload_functions.append(func)

    def register_cancel_obj(self, obj):
        """Add a ``weakref`` to an obj which has a ``cancel`` function.
        When the rule gets unloaded the cancel function will be called (if the obj was not already garbage collected)

        :param obj:
        """
        self.__cancel_objs.add(obj)

    # -----------------------------------------------------------------------------------------------------------------
    # deprecated functions
    # -----------------------------------------------------------------------------------------------------------------
    def item_watch(self, name: typing.Union[str, HABApp.core.items.BaseValueItem],
                   seconds_constant: int, watch_only_changes=True):
        warnings.warn("'item_watch' is deprecated, use the 'watch_change' or 'watch_update'"
                      " method of the item instance instead", DeprecationWarning, 2)

        assert isinstance(seconds_constant, int)
        assert isinstance(watch_only_changes, bool)
        if isinstance(name, str):
            item = HABApp.core.items.BaseValueItem.get_item(name)
        else:
            assert isinstance(name, HABApp.core.items.BaseValueItem), type(name)
            item = name

        if watch_only_changes:
            return item.watch_change(seconds_constant)
        else:
            return item.watch_update(seconds_constant)


    def item_watch_and_listen(self, name: typing.Union[HABApp.core.items.BaseValueItem, str],
                              seconds_constant: int, callback, watch_only_changes=True
                              ):
        warnings.warn("'item_watch_and_listen' is deprecated, use the 'watch_change' or 'watch_update'"
                      " method of the item instance instead", DeprecationWarning, 2)

        watched_item = self.item_watch(name, seconds_constant, watch_only_changes)
        event_listener = self.listen_event(
            name,
            callback,
            HABApp.core.events.ItemNoChangeEvent if watch_only_changes else HABApp.core.events.ItemNoUpdateEvent
        )
        return watched_item, event_listener


    def item_exists(self, name: str) -> bool:
        warnings.warn("'item_exists' is deprecated!", DeprecationWarning, 2)
        assert isinstance(name, str), type(name)
        return HABApp.core.Items.item_exists(name)

    def get_item_state(self, name: str, default=None) -> typing.Any:
        warnings.warn("'get_item_state' is deprecated, use 'get_value' or 'value'"
                      " method of the item instance instead", DeprecationWarning, 2)

        if default is None:
            return HABApp.core.Items.get_item(name).value

        try:
            state = HABApp.core.Items.get_item(name).value
        except HABApp.core.Items.ItemNotFoundException:
            return default

        if state is None:
            return default
        return state

    def set_item_state(self, name: str, value: typing.Any):
        warnings.warn("'set_item_state' is deprecated, use 'post_value' or 'set_value' method "
                      "of the item instance instead", DeprecationWarning, 2)

        if isinstance(name, str):
            try:
                item = HABApp.core.Items.get_item(name)
            except HABApp.core.Items.ItemNotFoundException:
                item = HABApp.core.Items.create_item(name, HABApp.core.items.Item)
        else:
            assert isinstance(name, HABApp.core.items.Item)
            item = name

        item.post_value(value)
        return None

    def get_item(self, name: str, item_factory=None) -> HABApp.core.items.Item:

        warnings.warn("'get_item' is deprecated, use 'Item.get_item' or 'Item.get_create_item' instead",
                      DeprecationWarning, 2)

        if item_factory is None:
            return HABApp.core.Items.get_item(name)

        try:
            return HABApp.core.Items.get_item(name)
        except HABApp.core.Items.ItemNotFoundException:
            return HABApp.core.Items.create_item(name, item_factory)

    def get_rule_parameter(self, file_name: str, *keys, default_value='ToDo'):
        warnings.warn("'get_rule_parameter' is deprecated, use 'HABApp.parameters.Parameter()' instead",
                      DeprecationWarning, 2)

        assert isinstance(file_name, str), type(file_name)
        import HABApp.parameters
        return HABApp.parameters.Parameter(file_name, *keys, default_value=default_value)

    # -----------------------------------------------------------------------------------------------------------------
    # internal functions
    # -----------------------------------------------------------------------------------------------------------------
    def _get_cb_name(self, callback):
        return f'{self.rule_name}.{callback.__name__}' if self.rule_name else None

    def _add_event_listener(self, listener: HABApp.core.EventBusListener) -> HABApp.core.EventBusListener:
        self.__event_listener.append(listener)
        HABApp.core.EventBus.add_listener(listener)
        return listener

    @HABApp.core.wrapper.log_exception
    def _check_rule(self):

        # Check if items do exists
        if not HABApp.core.Items.get_all_items():
            return None

        for listener in self.__event_listener:

            # Internal topics - don't warn there
            if listener.topic in HABApp.core.const.topics.ALL:
                continue

            # check if specific item exists
            if not HABApp.core.Items.item_exists(listener.topic):
                log.warning(f'Item "{listener.topic}" does not exist (yet)! '
                            f'self.listen_event in "{self.rule_name}" may not work as intended.')


    @HABApp.core.wrapper.log_exception
    def _process_events(self, now):

        # sheduled events
        clean_events = False
        for future_event in self.__future_events:  # type: OneTimeCallback
            future_event.check_due(now)
            future_event.execute()
            if future_event.is_finished:
                clean_events = True

        # remove finished events
        if clean_events:
            self.__future_events = [k for k in self.__future_events if not k.is_finished]
        return None

    @HABApp.core.wrapper.log_exception
    def _unload(self):

        # unload all functions
        for func in self.__unload_functions:
            try:
                func()
            except Exception as e:

                # try getting function name
                try:
                    name = f' in "{func.__name__}"'
                except AttributeError:
                    name = ''

                log.error(f'Error{name} while unloading "{self.rule_name}": {e}')

                # log traceback
                lines = traceback.format_exc().splitlines()
                del lines[1:3]  # see implementation in wrappedfunction.py why we do this
                for line in lines:
                    log.error(line)


@HABApp.core.wrapper.log_exception
def get_parent_rule() -> Rule:
    depth = 1
    while True:
        try:
            frm = sys._getframe(depth)
        except ValueError:
            raise RuntimeError('Could not find parent rule!') from None

        __vars = frm.f_locals
        depth += 1
        if 'self' in __vars:
            rule = __vars['self']
            if isinstance(rule, Rule):
                return rule



import logging
import logging.config
import time
import traceback
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler
from pathlib import Path

import ruamel.yaml

from HABApp import __version__
from . import CONFIG
from .default_logfile import get_default_logfile

_yaml_param = ruamel.yaml.YAML(typ='safe')
_yaml_param.default_flow_style = False
_yaml_param.default_style = False   # type: ignore
_yaml_param.width = 1000000         # type: ignore
_yaml_param.allow_unicode = True
_yaml_param.sort_base_mapping_type_on_output = False    # type: ignore


log = logging.getLogger('HABApp.Config')


class AbsolutePathExpected(Exception):
    pass


class InvalidConfigException(Exception):
    pass


class HABAppConfigLoader:

    def __init__(self, config_folder: Path):

        assert isinstance(config_folder, Path)
        assert config_folder.is_dir(), config_folder
        self.folder_conf = config_folder
        self.file_conf_habapp  = self.folder_conf / 'config.yml'
        self.file_conf_logging = self.folder_conf / 'logging.yml'

        # if the config does not exist it will be created
        self.__check_create_logging()

        # Load Config initially
        self.first_start = True
        try:
            # try load logging config first. If we use abs path we can log errors when loading config.yml
            self.on_file_event(self.file_conf_logging)
            self.on_file_event(self.file_conf_habapp)
        except AbsolutePathExpected:
            self.on_file_event(self.file_conf_habapp)
            self.on_file_event(self.file_conf_logging)
        self.first_start = False

    def on_file_event(self, path: Path):
        if path.name == 'config.yml':
            self.load_cfg()
        if path.name == 'logging.yml':
            self.load_log()

    def __check_create_logging(self):
        if self.file_conf_logging.is_file():
            return None

        print(f'Creating {self.file_conf_logging.name} in {self.file_conf_logging.parent}')
        with self.file_conf_logging.open('w', encoding='utf-8') as file:
            file.write(get_default_logfile())

        time.sleep(0.1)
        return None

    def load_cfg(self):

        CONFIG.load(self.file_conf_habapp)

        # check if folders exist and print warnings, maybe because of missing permissions
        if not CONFIG.directories.rules.is_dir():
            log.warning( f'Folder for rules files does not exist: {CONFIG.directories.rules}')

        log.debug('Loaded HABApp config')
        return None

    def load_log(self):
        # config gets created on startup - if it gets deleted we do nothing here
        if not self.file_conf_logging.is_file():
            return None

        with self.file_conf_logging.open('r', encoding='utf-8') as file:
            cfg = _yaml_param.load(file)

        # fix filenames
        for handler, handler_cfg in cfg.get('handlers', {}).items():

            # fix encoding for FileHandlers - we always log utf-8
            if 'file' in handler_cfg.get('class', '').lower():
                enc = handler_cfg.get('encoding', '')
                if enc != 'utf-8':
                    handler_cfg['encoding'] = 'utf-8'

            if 'filename' not in handler_cfg:
                continue

            # make Filenames absolute path in the log folder if not specified
            p = Path(handler_cfg['filename'])
            if not p.is_absolute():
                # Our log folder ist not yet converted to path -> it is not loaded
                if not isinstance(CONFIG.directories.logging, Path):
                    raise AbsolutePathExpected()

                # Use defined parent folder
                p = (CONFIG.directories.logging / p).resolve()
                handler_cfg['filename'] = str(p)

        # load prepared logging
        try:
            logging.config.dictConfig(cfg)
        except Exception as e:
            print(f'Error loading logging config: {e}')
            return None
        log.debug('Loaded logging config')

        # Try rotating the logs on first start
        if self.first_start:
            for handler in logging._handlerList:
                try:
                    handler = handler()  # weakref -> call it to get object
                    if isinstance(handler, (RotatingFileHandler, TimedRotatingFileHandler)):
                        handler.doRollover()
                except Exception:
                    lines = traceback.format_exc().splitlines()
                    # cut away AbsolutePathExpected Exception from log output
                    start = 0
                    for i, line in enumerate(lines):
                        if line.startswith('Traceback'):
                            start = i
                    for line in lines[start:]:
                        log.error(line)

        logging.getLogger('HABApp').info(f'HABApp Version {__version__}')
        return None



import re
import typing
from dataclasses import dataclass
from typing import Dict, List
from typing import Iterator, Optional, Union

from pydantic import BaseModel, Field, ValidationError, parse_obj_as, validator

from HABApp.core.logger import HABAppError
from HABApp.openhab.connection_logic.plugin_things.filters import ChannelFilter, ThingFilter
from HABApp.openhab.connection_logic.plugin_things.str_builder import StrBuilder
from HABApp.openhab.definitions.definitions import ITEM_TYPES
from ._log import log

RE_VALID_NAME = re.compile(r'\w+')


@dataclass
class UserItem:
    type: str
    name: str
    label: str
    icon: str
    groups: List[str]
    tags: List[str]
    link: Optional[str]
    metadata: Dict[str, Dict[str, Union[str, int, float]]]

    def get_oh_cfg(self) -> Dict[str, Union[str, dict, list]]:
        ret = {}
        for k in self.__annotations__:
            if k in ('link', 'metadata'):
                continue

            v = self.__dict__[k]
            if k == 'icon':
                k = 'category'
            ret[k] = v
        return ret


class InvalidItemNameError(Exception):
    pass


class MetadataCfg(BaseModel):
    value: str
    config: Dict[str, typing.Any] = {}


class UserItemCfg(BaseModel):
    type: str
    name: str
    label: str = ''
    icon: str = ''
    groups: List[str] = []
    tags: List[str] = []
    metadata: Optional[Dict[str, MetadataCfg]] = None

    @validator('type', always=True)
    def validate_item_type(cls, v):
        if v in ITEM_TYPES:
            return v
        try:
            return {k.lower(): k for k in ITEM_TYPES}[v.lower()]
        except KeyError:
            raise ValueError(f'Must be one of {", ".join(ITEM_TYPES)}')

    @validator('name', 'label', 'icon', 'groups', 'tags', each_item=True, always=True)
    def validate_make_str_builder(cls, v):
        return StrBuilder(v)

    @validator('metadata', pre=True)
    def make_meta_cfg(cls, v):
        if not isinstance(v, dict):
            return v

        for key, val in v.items():
            if isinstance(val, str):
                v[key] = {'value': val}
        return v

    def get_item(self, context: dict) -> UserItem:
        v = {'link': None}
        for k in self.__fields__:
            val = self.__dict__[k]

            # type is const
            if k == 'type':
                v[k] = val
                continue

            # metadata is nested
            if k == 'metadata':
                v[k] = {k: v.dict() for k, v in val.items()} if val is not None else {}
                continue

            # resolve str wildcards
            if k in ('groups', 'tags'):
                v[k] = [s.get_str(context) for s in val]
                continue

            v[k] = val.get_str(context)

        # ensure a valid item name, otherwise the creation will definitely fail
        v['name'] = name = v['name'].replace('ä', 'ae').replace('ö', 'oe').replace('ü', 'ue').replace(' ', '_')
        if not RE_VALID_NAME.fullmatch(name):
            raise InvalidItemNameError(f'"{name}" is not a valid name for an item!\n   (created for {context})')
        return UserItem(**v)


class UserChannelCfg(BaseModel):
    filter: List[ChannelFilter]
    link_items: List[UserItemCfg] = Field(default_factory=list, alias='link items')

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True

    @validator('filter', pre=True, always=True)
    def validate_filter(cls, v):
        return create_filters(ChannelFilter, v)

    def get_items(self, context: dict) -> Iterator[UserItem]:
        return map(lambda x: x.get_item(context), self.link_items)


class UserThingCfg(BaseModel):
    test: bool
    filter: List[ThingFilter]
    # order of the type hint matters: int, str!
    thing_config: Dict[Union[int, str], Union[int, float, str]] = Field(alias='thing config', default_factory=dict)
    create_items: List[UserItemCfg] = Field(alias='create items', default_factory=list)
    channels: List[UserChannelCfg] = Field(default_factory=list)

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True

    @validator('filter', pre=True, always=True)
    def validate_filter(cls, v):
        return create_filters(ThingFilter, v)

    def get_items(self, context: dict) -> Iterator[UserItem]:
        return map(lambda x: x.get_item(context), self.create_items)


def create_filters(cls, v: Union[List[Dict[str, str]], Dict[str, str]]):
    if isinstance(v, dict):
        v = [v]
    r = []
    for a in v:
        if not isinstance(a, dict):
            raise ValueError(f'Entry {a} is not a valid dict!')
        for key, regex in a.items():
            r.append(cls(key, regex))
    return r


def validate_cfg(_in, filename: Optional[str] = None) -> Optional[List[UserThingCfg]]:
    try:
        if isinstance(_in, list):
            return parse_obj_as(List[UserThingCfg], _in, type_name=filename)
        else:
            return [parse_obj_as(UserThingCfg, _in, type_name=filename)]
    except ValidationError as e:
        HABAppError(log).add_exception(e).dump()
        return None



import asyncio
import datetime
import logging
import typing
from datetime import timedelta

from HABApp.core.wrapper import log_exception
from .base_item_watch import BaseWatch, ItemNoChangeWatch, ItemNoUpdateWatch
from ..const import loop

log = logging.getLogger('HABApp')


class ItemTimes:
    WATCH: typing.Union[typing.Type[ItemNoUpdateWatch], typing.Type[ItemNoChangeWatch]]

    def __init__(self, name: str, dt: datetime.datetime):
        self.name: str = name
        self.dt: datetime.datetime = dt
        self.tasks: typing.List[BaseWatch] = []

    def set(self, dt: datetime.datetime, events=True):
        self.dt = dt
        if not self.tasks:
            return

        if events:
            asyncio.run_coroutine_threadsafe(self.schedule_events(), loop)
        return None

    def add_watch(self, secs: typing.Union[int, float, timedelta]) -> BaseWatch:
        if isinstance(secs, timedelta):
            secs = secs.total_seconds()
        assert secs > 0, secs

        # don't add the watch two times
        for t in self.tasks:
            if not t.fut.is_canceled and t.fut.secs == secs:
                log.warning(f'Watcher {self.WATCH.__name__} ({t.fut.secs}s) for {self.name} has already been created')
                return t

        w = self.WATCH(self.name, secs)
        self.tasks.append(w)
        log.debug(f'Added {self.WATCH.__name__} ({w.fut.secs}s) for {self.name}')
        return w

    @log_exception
    async def schedule_events(self):
        canceled = []
        for t in self.tasks:
            if t.fut.is_canceled:
                canceled.append(t)
            else:
                t.fut.reset()

        # remove canceled tasks
        if canceled:
            for c in canceled:
                self.tasks.remove(c)
                log.debug(f'Removed {self.WATCH.__name__} ({c.fut.secs}s) for {self.name}')
        return None


class UpdatedTime(ItemTimes):
    WATCH = ItemNoUpdateWatch


class ChangedTime(ItemTimes):
    WATCH = ItemNoChangeWatch
